# SPDX-License-Identifier: BSD-2-Clause
""" Converts Doxygen XML content to interface specification files. """

# Copyright (C) 2024, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dataclasses
import os
import re
from typing import Any, Iterator
from xml.etree import ElementTree


@dataclasses.dataclass
class DoxygenContext:
    """ Represents the Doxygen context. """
    config: dict
    items_by_kind: dict[str, dict[str, "DoxygenItem"]] = dataclasses.field(
        default_factory=dict)
    items_by_name: dict[str, dict[str,
                                  list["DoxygenItem"]]] = dataclasses.field(
                                      default_factory=dict)
    items: dict[str, "DoxygenItem"] = dataclasses.field(default_factory=dict)
    compound_typedefs: dict[str, str] = dataclasses.field(default_factory=dict)


_Data = dict[str, Any]

_INVALID_NAME_CHARS = re.compile(r"[^a-zA-Z0-9]+")

_FUNCTION_POINTER = re.compile(r"([^(]+)\(\*\)\((.*)")


def _strip(text: str | None) -> str | None:
    if text is None:
        return None
    text = text.strip()
    if not text:
        return None
    return text


def _decl(defs: dict[str, str], index: int) -> str:
    name = f"${{.:/params[{index}]/name}}"
    type_name = defs.get("type", None)
    if type_name is None:
        return name
    mobj = _FUNCTION_POINTER.match(type_name)
    if mobj is not None:
        type_name = mobj.group(1).strip()
        name = f"(*{name})({mobj.group(2)}"
    if type_name.endswith("*"):
        return f"{type_name}{name}"
    return f"{type_name} {name}"


@dataclasses.dataclass
class DoxygenItem:
    """ Represents a Doxygen item. """
    ctx: DoxygenContext
    kind: str
    doxygen_id: str
    name: str
    data: _Data = dataclasses.field(default_factory=lambda: {
        "brief": "",
        "description": ""
    })
    group_ids: list[str] = dataclasses.field(default_factory=list)
    file_ids: set[str] = dataclasses.field(default_factory=set)

    def __lt__(self, other: "DoxygenItem") -> bool:
        return self.doxygen_id < other.doxygen_id

    def __str__(self) -> str:
        fields = ", ".join(f"{field.name}={getattr(self, field.name)!r}"
                           for field in dataclasses.fields(self)
                           if field.name != "ctx")
        return f"{type(self).__name__}({fields})"

    @property
    def uid(self) -> str:
        """ Is the UID of the item. """
        group = self.group
        prefix = os.path.dirname(group.uid)
        name = self.name.lower()
        name = _INVALID_NAME_CHARS.sub("-", name)
        name = name.removeprefix(self.ctx.config["groups"].get(
            group.doxygen_id, {}).get("remove-prefix", ""))
        return os.path.join(prefix, name)

    def uid_relative_to(self, other: str) -> str:
        """ Get the UID of the item relative to the other UID. """
        return os.path.relpath(self.uid, os.path.dirname(other))

    @property
    def file(self) -> "DoxygenFile":
        """ Is the first file of the item. """
        try:
            return next(self.files)
        except StopIteration as err:
            raise ValueError(f"{self.name} has no file: {self}") from err

    @property
    def files(self) -> Iterator["DoxygenFile"]:
        """ Is the files of the item. """
        for file_id in self.file_ids:
            file = self.ctx.items[file_id]
            assert isinstance(file, DoxygenFile)
            yield file

    @property
    def group(self) -> "DoxygenGroup":
        """ Is the first group of the item. """
        try:
            return next(self.groups)
        except StopIteration as err:
            raise ValueError(f"{self.name} has no group: {self}") from err

    @property
    def groups(self) -> Iterator["DoxygenGroup"]:
        """ Is the group set of the item. """
        for group_id in self.group_ids:
            group = self.ctx.items[group_id]
            assert isinstance(group, DoxygenGroup)
            yield group

    @property
    def is_header(self) -> bool:
        """ Indicates if the item is a header file. """
        return False

    def _get_optional_text(self, key: str) -> str | None:
        return _strip(self.data.get(key, None))

    @property
    def brief(self) -> str | None:
        """ Is the brief description of the item. """
        return self._get_optional_text("brief")

    @property
    def description(self) -> str | None:
        """ Is the detailed description of the item. """
        return self._get_optional_text("description")

    @property
    def notes(self) -> str | None:
        """ Is the notes section of the item. """
        return self._get_optional_text("notes")

    def export(self) -> dict:
        """ Export the Doxygen item as specification item data. """
        links: list[dict] = []
        for file in self.files:
            if file.is_header:
                links.append({
                    "role": "interface-placement",
                    "uid": file.uid_relative_to(self.uid)
                })
        for group in self.groups:
            links.append({
                "role": "interface-ingroup",
                "uid": group.uid_relative_to(self.uid)
            })
        data = {
            "SPDX-License-Identifier": "CC-BY-SA-4.0 OR BSD-2-Clause",
            "brief": self.brief,
            "copyrights": [],
            "description": self.description,
            "enabled-by": True,
            "index-entries": [],
            "links": links,
            "notes": self.notes,
            "type": "interface"
        }
        data.update(self.ctx.config["data"])
        return data

    def _get_initializer(self) -> str | None:
        body = self.data.get("initializer", None)
        if body is None:
            return None
        return body.replace("&gt;", ">").replace("&lt;", "<")

    def add_function_like_attributes(self, interface_type: str,
                                     data: dict[str, Any]) -> None:
        """ Add function-like attributes to the data. """
        data["name"] = self.name
        data["interface-type"] = interface_type
        params: list[dict] = []
        paramdefs: list[str] = []
        if len(self.data["param"]) == len(self.data["paramdefs"]):
            for index, (param, definition) in enumerate(
                    zip(self.data["param"], self.data["paramdefs"])):
                params.append({
                    "description": param["description"].strip(),
                    "dir": param["dir"],
                    "name": param["name"]
                })
                paramdefs.append(_decl(definition, index))
        else:
            for index, definition in enumerate(self.data["paramdefs"]):
                if definition["type"] == "void":
                    continue
                params.append({
                    "description": None,
                    "dir": None,
                    "name": definition["declname"]
                })
                paramdefs.append(_decl(definition, index))
        data["params"] = params
        type_name = self.data.get("type", None)
        data["definition"] = {
            "default": {
                "attributes": None,
                "body": self._get_initializer(),
                "params": paramdefs,
                "return": None if type_name == "void" else type_name
            },
            "variants": []
        }
        retval = self.data.get("retval", [])
        if "return" in self.data or retval:
            data["return"] = {
                "return":
                _strip(self.data.get("return", None)),
                "return-values": [{
                    "description": info["description"].strip(),
                    "value": info["name"]
                } for info in retval]
            }
        else:
            data["return"] = None

    def add_definition_kind(self, data: dict[str, Any]) -> None:
        """ Adds the definition kind attribute to the data. """
        if self.doxygen_id in self.ctx.compound_typedefs:
            if self.ctx.compound_typedefs[self.doxygen_id] is None:
                data["definition-kind"] = "typedef-only"
            else:
                data["definition-kind"] = f"typedef-and-{self.kind}"
        else:
            data["definition-kind"] = f"{self.kind}-only"


@dataclasses.dataclass
class DoxygenContainer(DoxygenItem):
    """ Represents a Doxygen container item. """
    member_ids: list[str] = dataclasses.field(default_factory=list)

    def members(self) -> Iterator[DoxygenItem]:
        """ Yields the members of the item. """
        for member_id in self.member_ids:
            yield self.ctx.items[member_id]


@dataclasses.dataclass
class DoxygenCompound(DoxygenContainer):
    """ Represents a Doxygen compound item. """

    def export(self) -> dict:
        data = super().export()
        data["name"] = self.name
        data["interface-type"] = self.kind
        self.add_definition_kind(data)
        definition: list[dict] = []
        for member in self.members():
            definition.append({
                "default": {
                    "brief":
                    member.brief,
                    "definition":
                    member.data["definition"].replace(
                        f"{self.name}::{member.name}", "${.:name}"),
                    "description":
                    member.description,
                    "kind":
                    "member",
                    "name":
                    member.name
                },
                "variants": []
            })
        data["definition"] = definition
        return data


@dataclasses.dataclass
class DoxygenDefine(DoxygenItem):
    """ Represents a Doxygen define item. """

    def export(self) -> dict:
        data = super().export()
        if self.data.get("param", []):
            self.add_function_like_attributes("macro", data)
        else:
            data["name"] = self.name
            data["interface-type"] = "define"
            data["definition"] = {
                "default": self._get_initializer(),
                "variants": []
            }
        return data


@dataclasses.dataclass
class DoxygenDir(DoxygenItem):
    """ Represents a Doxygen directory item. """


@dataclasses.dataclass
class DoxygenEnum(DoxygenContainer):
    """ Represents a Doxygen enumeration item. """

    def export(self) -> dict:
        data = super().export()
        data["name"] = self.name
        data["interface-type"] = "enum"
        self.add_definition_kind(data)
        for member in self.members():
            data["links"].append({
                "role": "interface-enumerator",
                "uid": member.uid_relative_to(self.uid)
            })
        return data


@dataclasses.dataclass
class DoxygenEnumValue(DoxygenItem):
    """ Represents a Doxygen enumerator item. """

    def export(self) -> dict:
        data = super().export()
        data["name"] = self.name
        data["interface-type"] = "enumerator"
        data["links"] = []
        return data


@dataclasses.dataclass
class DoxygenFile(DoxygenContainer):
    """ Represents a Doxygen file item. """

    @property
    def uid(self) -> str:
        the_uid = super().uid
        if not self.is_header:
            return the_uid
        basename = os.path.basename(the_uid)[:-2]
        if basename:
            basename = f"header-{basename}"
        else:
            basename = "header"
        return os.path.join(os.path.dirname(the_uid), basename)

    @property
    def is_header(self) -> bool:
        return self.name.endswith(".h")


@dataclasses.dataclass
class DoxygenFunction(DoxygenItem):
    """ Represents a Doxygen function item. """

    def export(self) -> dict:
        data = super().export()
        self.add_function_like_attributes("function", data)
        return data


@dataclasses.dataclass
class DoxygenGroup(DoxygenContainer):
    """ Represents a Doxygen group item. """

    @property
    def uid(self) -> str:
        return self.ctx.config["groups"].get(self.doxygen_id,
                                             {}).get("uid",
                                                     f"/{self.name.lower()}")


@dataclasses.dataclass
class DoxygenStruct(DoxygenCompound):
    """ Represents a Doxygen structure item. """


@dataclasses.dataclass
class DoxygenTypedef(DoxygenItem):
    """ Represents a Doxygen typedef item. """


@dataclasses.dataclass
class DoxygenUnion(DoxygenCompound):
    """ Represents a Doxygen union item. """


@dataclasses.dataclass
class DoxygenVariable(DoxygenItem):
    """ Represents a Doxygen variable item. """


_ITEM_TYPES = {
    "define": DoxygenDefine,
    "dir": DoxygenDir,
    "enum": DoxygenEnum,
    "enumvalue": DoxygenEnumValue,
    "file": DoxygenFile,
    "function": DoxygenFunction,
    "group": DoxygenGroup,
    "struct": DoxygenStruct,
    "typedef": DoxygenTypedef,
    "union": DoxygenUnion,
    "variable": DoxygenVariable
}


@dataclasses.dataclass
class _Scope:
    item: DoxygenItem
    data: Any
    key: str


def _tag_attribute(elem: ElementTree.Element, scope: _Scope,
                   attribute: str) -> _Scope:
    text = elem.text
    scope.data[attribute] = text if text is not None else ""
    return _Scope(scope.item, scope.data, attribute)


def _tag_item(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    item = scope.item.ctx.items[_doxygen_id(elem)]
    if item.kind == "variable" and isinstance(scope.item, DoxygenCompound):
        scope.item.member_ids.append(item.doxygen_id)
    elif item.kind == "enumvalue":
        enum = scope.item
        assert isinstance(enum, DoxygenEnum)
        enum.member_ids.append(item.doxygen_id)

    # Doxygen places the description into multiple files.  Lets hope they are
    # all the same.
    item.data["brief"] = ""
    item.data["description"] = ""
    item.data["param"] = []
    item.data["paramdefs"] = []

    return _Scope(item, item.data, "")


def _tag_add_text(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    text = elem.text
    if text is not None:
        scope.data[scope.key] = f"{scope.data[scope.key]} {text} "
    return scope


def _tag_brief(_elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _Scope(scope.item, scope.data, "brief")


def _tag_description(_elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _Scope(scope.item, scope.data, "description")


def _tag_param(_elem: ElementTree.Element, scope: _Scope) -> _Scope:
    paramdef: dict[str, str] = {}
    scope.item.data["paramdefs"].append(paramdef)
    return _Scope(scope.item, paramdef, "")


def _tag_parameteritem(_elem: ElementTree.Element, scope: _Scope) -> _Scope:
    param = {"description": ""}
    scope.item.data[scope.key].append(param)
    return _Scope(scope.item, param, "description")


def _tag_parameterlist(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    kind = elem.attrib["kind"]
    scope.item.data[kind] = []
    return _Scope(scope.item, scope.data, kind)


def _tag_parametername(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    scope.data["name"] = elem.text
    try:
        direction = elem.attrib["direction"]
    except KeyError:
        direction = None
    scope.data["dir"] = direction
    return scope


def _tag_ref(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    text = elem.text
    assert text is not None
    tail = elem.tail
    if tail is not None:
        text += tail
    scope.data[scope.key] += text
    return scope


def _tag_type(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, "type")


def _tag_definition(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, "definition")


def _tag_argsstring(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, "argsstring")


def _tag_initializer(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, "initializer")


def _tag_declname(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, "declname")


def _tag_simplesect(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, elem.attrib["kind"])


def _tag_title(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    return _tag_attribute(elem, scope, "title")


def _tag_sectiondef(elem: ElementTree.Element, scope: _Scope) -> _Scope:
    kind = elem.attrib["kind"]
    more = {"description": ""}
    scope.data[f"section-{kind}"] = more
    return _Scope(scope.item, more, "description")


def _tag_listitem(_elem: ElementTree.Element, scope: _Scope) -> _Scope:
    scope.data[scope.key] = f"{scope.data[scope.key]}\n\n* "
    return scope


_TAG_HANDLER = {
    "argsstring": _tag_argsstring,
    "briefdescription": _tag_brief,
    "compounddef": _tag_item,
    "declname": _tag_declname,
    "definition": _tag_definition,
    "description": _tag_description,
    "detaileddescription": _tag_description,
    "enumvalue": _tag_item,
    "initializer": _tag_initializer,
    "listitem": _tag_listitem,
    "memberdef": _tag_item,
    "para": _tag_add_text,
    "param": _tag_param,
    "parameterdescription": _tag_description,
    "parameteritem": _tag_parameteritem,
    "parameterlist": _tag_parameterlist,
    "parametername": _tag_parametername,
    "ref": _tag_ref,
    "sectiondef": _tag_sectiondef,
    "simplesect": _tag_simplesect,
    "title": _tag_title,
    "type": _tag_type,
}


def _doxygen_id(elem: ElementTree.Element) -> str:
    try:
        return elem.attrib["id"]
    except KeyError:
        return elem.attrib["refid"]


def _relationships(elem: ElementTree.Element, item: DoxygenItem) -> None:
    assert isinstance(item, DoxygenContainer)
    for member_kind in ("enumvalue", "innerfile", "innerclass", "member",
                        "memberdef"):
        for member in elem.findall(f".//{member_kind}"):
            item.member_ids.append(_doxygen_id(member))


_RELATIONSHIP_HANDLER = {"file": _relationships, "group": _relationships}

_TAG_IS_KIND = {"enumvalue"}


def _get_kind_name(elem: ElementTree.Element) -> tuple[str, str]:
    try:
        kind = elem.attrib["kind"]
    except KeyError:
        if elem.tag not in _TAG_IS_KIND:
            raise
        kind = elem.tag
    name_tag = elem.find("name")
    if name_tag is None:
        name_tag = elem.find("compoundname")
    assert name_tag is not None
    name = name_tag.text
    assert name is not None
    return kind, name


_IGNORE = ("programlisting", "preformatted")


def _fill_items(elem: ElementTree.Element, scope: _Scope) -> None:
    if elem.tag in _IGNORE:
        return
    handler = _TAG_HANDLER.get(elem.tag, None)
    if handler is not None:
        scope = handler(elem, scope)
    for child in elem.findall("*"):
        _fill_items(child, scope)


_COMPOUND_TYPEDEF = re.compile(
    r"typedef\s+(enum|struct|union)(\s+[a-zA-Z0-9_]+)?")


def _learn_compound_typedefs(ctx: DoxygenContext,
                             elem: ElementTree.Element) -> None:
    text = " ".join(elem.itertext()).strip()
    mobj = _COMPOUND_TYPEDEF.match(text)
    if mobj:
        ctx.compound_typedefs[elem.attrib["refid"]] = mobj.group(2)


def _gather_doxygen_id_to_item_mappings(ctx: DoxygenContext,
                                        xml_files: list[str]):
    for xml_file in xml_files:
        tree = ElementTree.parse(xml_file)
        for elem in tree.iter():
            if elem.tag == "codeline" and elem.attrib.get(
                    "refkind", None) in ("compound", "member"):
                _learn_compound_typedefs(ctx, elem)
                continue
            try:
                doxygen_id = elem.attrib["id"]
            except KeyError:
                continue
            try:
                kind, name = _get_kind_name(elem)
            except KeyError:
                continue
            item = ctx.items.setdefault(
                doxygen_id, _ITEM_TYPES[kind](ctx, kind, doxygen_id, name))
            ctx.items_by_kind.setdefault(kind, {}).setdefault(doxygen_id, item)
            ctx.items_by_name.setdefault(kind, {}).setdefault(name,
                                                              []).append(item)
            handler = _RELATIONSHIP_HANDLER.get(kind, None)
            if handler is not None:
                handler(elem, item)


def _add_group_associations(ctx: DoxygenContext):
    for group in ctx.items_by_kind["group"].values():
        assert isinstance(group, DoxygenGroup)
        group_id = group.doxygen_id
        for member_id in group.member_ids:
            ctx.items[member_id].group_ids.append(group_id)


def _add_file_associations(ctx: DoxygenContext):
    for file in ctx.items_by_kind["file"].values():
        assert isinstance(file, DoxygenFile)
        for member_id in file.member_ids:
            ctx.items[member_id].file_ids.add(file.doxygen_id)


def _add_group_through_file_associations(ctx: DoxygenContext):
    for kind, items in ctx.items_by_kind.items():
        if kind in ("dir", "group"):
            continue
        for item in items.values():
            if not item.group_ids:
                try:
                    file_item = ctx.items[item.file.doxygen_id]
                    assert isinstance(file_item, DoxygenFile)
                except ValueError:
                    pass
                else:
                    try:
                        item.group_ids.append(file_item.group_ids[0])
                    except IndexError:
                        pass


def doxygen_xml_to_spec(config: dict, xml_files: list[str]) -> DoxygenContext:
    """ Convert Doxygen XML files to specification item data.  """
    ctx = DoxygenContext(config)

    # In the first pass get the Doxygen identifier to item mappings and vice
    # versa.  Associate items with groups.
    _gather_doxygen_id_to_item_mappings(ctx, xml_files)
    _add_group_associations(ctx)
    _add_file_associations(ctx)
    _add_group_through_file_associations(ctx)

    # In the second pass fill the items
    for xml_file in xml_files:
        tree = ElementTree.parse(xml_file)
        _fill_items(tree.getroot(),
                    _Scope(DoxygenItem(ctx, "root", "", ""), {}, ""))

    return ctx



# SPDX-License-Identifier: BSD-2-Clause
""" Produces GCDA files from a test log. """

# Copyright (C) 2022 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import glob
import json
import logging
import os
import shutil
from subprocess import run as subprocess_run

from .directorystate import DirectoryState


class GCDAProducer(DirectoryState):
    """ Runs the gcov-tool to produce GCDA files from a test log. """

    def run(self):
        super().run()
        self.discard()

        build = self.input("build")
        assert isinstance(build, DirectoryState)

        logging.info("%s: copy *.gcno files from '%s' to '%s'", self.uid,
                     build.directory, self.directory)
        for file in build.files():
            assert not file.endswith(".gcda")
            if file.endswith(".gcno"):
                file_dest = file.replace(build.directory, self.directory)
                os.makedirs(os.path.dirname(file_dest), exist_ok=True)
                shutil.copy2(file, file_dest)

        gcda_pattern = f"{build.directory}/**/*.gcda"
        for file in glob.glob(gcda_pattern, recursive=True):
            logging.warning(
                "%s: remove unexpected *.gcda file in build directory: '%s'",
                self.uid, file)
            os.remove(file)

        log = self.input("log")
        assert isinstance(log, DirectoryState)

        gcov_tool = self["gcov-tool"]
        cwd = self["working-directory"]

        logging.debug("%s: load file: %s", self.uid, log.file)
        with open(log.file, "r", encoding="utf-8") as src:
            data = json.load(src)

        for report in data["reports"]:
            logging.debug("%s: consider: %s", self.uid, report["executable"])
            if "line-end-of-test" not in report["info"]:
                logging.info("%s: discard coverage data of failed test: %s",
                             self.uid, report["executable"])
                continue
            begin = report.get("line-gcov-info-base64-begin", None)
            if begin is None:
                logging.info("%s: discard due to missing gcov info begin: %s",
                             self.uid, report["executable"])
                continue
            end = report.get("line-gcov-info-base64-end", None)
            if end is None:
                logging.info("%s: discard due to missing gcov info end: %s",
                             self.uid, report["executable"])
                continue
            if report.get("gcov-info-hash",
                          "") != report.get("gcov-info-hash-calculated", ""):
                logging.info("%s: discard corrupt report: %s", self.uid,
                             report["executable"])
                continue
            logging.debug("%s: process: %s", self.uid, report["executable"])
            gcov_info = base64.b64decode("".join(report["output"][begin +
                                                                  1:end]))
            subprocess_run([gcov_tool, "merge-stream"],
                           check=True,
                           cwd=cwd,
                           input=gcov_info)

        logging.info("%s: move *.gcda files from '%s' to '%s'", self.uid,
                     build.directory, self.directory)
        for file in glob.glob(gcda_pattern, recursive=True):
            file_dest = file.replace(build.directory, self.directory)
            os.replace(file, file_dest)

        self.create_symbolic_links(self["symbolic-links"])
        self.description.add(f"""Produce GCDA files in directory
{self.description.path(self.directory)}""")
        self.run_post_actions()



# SPDX-License-Identifier: BSD-2-Clause
""" Run tests. """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json
import logging
import math
import multiprocessing
import os
import re
from pathlib import Path
import queue
import subprocess
from subprocess import run as subprocess_run
import tarfile
import time
import threading
from typing import Any, NamedTuple

from specitems import is_enabled, Item, ItemGetValueContext

from .directorystate import DirectoryState
from .pkgitems import BuildItem, PackageBuildDirector
from .testoutputparser import augment_report
from .util import now_utc

Report = dict[str, Any]


class Executable(NamedTuple):
    """ Represents a test executable. """
    path: str
    digest: str
    timeout: float


class TestLog(DirectoryState):
    """ Maintains a test log. """

    def get_reports_by_hash(self) -> tuple[dict[str, Report], str, str]:
        """
        Get the reports by executable hash and the test runner description.
        """
        reports_by_hash: dict[str, Report] = {}
        description = "There is no description available."
        runner_hash = ""
        try:
            with open(self.file, "r", encoding="utf-8") as src:
                data = json.load(src)
                description = data.get("test-runner-description", description)
                runner_hash = data.get("test-runner-hash", runner_hash)
                for report in data["reports"]:
                    digest = report["executable-sha512"]
                    assert digest not in reports_by_hash
                    assert isinstance(digest, str)
                    reports_by_hash[digest] = report
        except FileNotFoundError:
            pass
        return reports_by_hash, description, runner_hash


class TestRunner(BuildItem):
    """ Runs tests. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self._executable = "/dev/null"
        self._executables: list[Executable] = []
        self.mapper.add_get_value(f"{self.item.type}:/test-executable",
                                  self._get_test_executable)

    def _get_test_executable(self, _ctx: ItemGetValueContext) -> str:
        return self._executable

    def describe(self) -> str:
        """ Return a description of the test runner. """
        return "This test procedure cannot run tests."

    def get_runner_hash(self) -> str:
        """
        Get a hash of the test runner.

        The runner hash is stored in the test log.  If the current runner hash
        differs from the hash in the test log, the test reports are not reused
        and all tests are run again.
        """
        return self.digest()

    def run_tests(self, executables: list[Executable]) -> list[Report]:
        """
        Run the test executables and returns a list of test reports.
        """
        start_time = now_utc()
        return [{
            "executable": executable.path,
            "executable-sha512": executable.digest,
            "command-line": "",
            "start-time": start_time,
            "output": ["This executable did not run."],
            "duration": 0.0
        } for executable in executables]

    def get_run_command(self, _executable: str) -> None | list[str]:
        """ Get the command to run an executable. """
        return None

    def _get_config_and_timeout_keys(self) -> tuple[str, str]:
        config = self.input("build-configuration")

        # Get timeout key for the enabled set
        for timeout_key in config["test-timeout-key"]:
            if is_enabled(self.enabled_set, timeout_key["enabled-by"]):
                return config["config-key"], timeout_key["key"]

        raise ValueError(
            f"{self.uid}: for enabled set {list(self.enabled_set)} "
            f"and build configuration {config.uid}, there is no timeout key "
            f"in {config['test-timeout-key']}")

    def _get_reports_and_executables(
            self, target: Item, previous_reports_by_hash: dict[str, Report],
            report_runner_hash: str,
            timeout_key: str) -> tuple[list[Report], list[Executable]]:
        # pylint: disable=too-many-locals
        timeouts = target.parent("test-timeouts")["timeouts"]
        source = self.input("source")
        assert isinstance(source, DirectoryState)
        reports: list[Report] = []
        executables: list[Executable] = []
        do_not_run = self["do-not-run"]
        default_timeout = self["default-timeout-in-seconds"]
        min_timeout = self["min-timeout-in-seconds"]
        timeout_scaler = self["timeout-scaler"]
        cannot_reuse_reports = self.get_runner_hash() != report_runner_hash
        if cannot_reuse_reports:
            logging.info(
                "%s: cannot reuse reports with "
                "runner hash %s and report hash %s", self.uid,
                self.get_runner_hash(), report_runner_hash)
        for path, digest in source.files_and_hashes():
            if not path.endswith(".exe") or path.endswith(".norun.exe"):
                continue
            assert digest
            name = Path(path).name

            # Use previous report if the executable hash did not change.  Reuse
            # also reports which contain an "error" attribute to avoid running
            # them in CI jobs.  Resolving failed tests is a manual activity.
            report = previous_reports_by_hash.get(digest, None)
            if report is None or cannot_reuse_reports:
                if name in do_not_run:
                    logging.info("%s: do not run: %s", self.uid, path)
                    report = TestRunner.run_tests(
                        self, [Executable(path, digest, 0.0)])[0]
                    augment_report(report, report["output"])
                    reports.append(report)
                else:
                    logging.debug("%s: run: %s", self.uid, path)
                    try:
                        timeout = max(timeouts[timeout_key][name])
                    except KeyError:
                        timeout = default_timeout
                    timeout = timeout_scaler * max(timeout, min_timeout)
                    executables.append(Executable(path, digest, timeout))
            else:
                logging.info("%s: use previous report for: %s", self.uid, path)
                report["executable"] = path
                reports.append(report)
        return reports, executables

    def run(self) -> None:
        start_time = now_utc()
        begin = time.monotonic()
        log = self.output("log")
        assert isinstance(log, TestLog)
        target = self.item.parent("target")
        assert target.uid == self.input("target").uid
        config_key, timeout_key = self._get_config_and_timeout_keys()
        (reports_by_hash, description,
         report_runner_hash) = log.get_reports_by_hash()
        reports, executables = self._get_reports_and_executables(
            target, reports_by_hash, report_runner_hash, timeout_key)

        # Run the tests with changed executable hashes
        if executables:
            description = self.describe()
            reports.extend(self._run_tests(executables))

        # Save the reports
        os.makedirs(os.path.dirname(log.file), exist_ok=True)
        with open(log.file, "w", encoding="utf-8") as dst:
            data = {
                "build-configuration": config_key,
                "timeout-key": timeout_key,
                "duration": time.monotonic() - begin,
                "end-time": now_utc(),
                "reports": sorted(reports, key=lambda x: x["executable"]),
                "start-time": start_time,
                "test-runner-description": description,
                "test-runner-hash": self.get_runner_hash(),
                "target": target.uid
            }
            json.dump(data, dst, sort_keys=True, indent=2)

        self.description.add(f"""Run tests and produce test result file
{self.description.path(log.file)}""")

    def _run_tests(self, executables: list[Executable]) -> list[Report]:
        max_run_count = self["max-retry-count-per-executable"] + 1
        reports_by_path: dict[str, Report] = {}
        while executables and max_run_count:
            self._executables = executables
            for new_report in self.run_tests(executables):
                augment_report(new_report, new_report["output"])
                reports_by_path[new_report["executable"]] = new_report
            next_executables: list[Executable] = []
            for executable in executables:
                report = reports_by_path.get(executable.path, None)
                if report is None:
                    next_executables.append(executable)
                    logging.warning("%s: executable '%s': no report", self.uid,
                                    executable.path)
                    continue
                if "error" in report:
                    next_executables.append(executable)
                    logging.warning("%s: executable '%s': %s", self.uid,
                                    executable.path, report["error"])
                    continue
                if report.get("gcov-info-hash",
                              "") != report.get("gcov-info-hash-calculated",
                                                ""):
                    next_executables.append(executable)
                    logging.warning(
                        "%s: executable '%s': gcov info is corrupt", self.uid,
                        executable.path)
                    continue
                test_suite = report.get("test-suite", {})
                if test_suite.get("report-hash", "") != test_suite.get(
                        "report-hash-calculated", ""):
                    next_executables.append(executable)
                    logging.warning(
                        "%s: executable '%s': test suite report is corrupt",
                        self.uid, executable.path)
                    continue
            executables = next_executables
            max_run_count -= 1
        return list(reports_by_path.values())


class GRMONManualTestRunner(TestRunner):
    """ Provides scripts to run the tests using GRMON. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.mapper.add_get_value(f"{self.item.type}:/test-executables-grmon",
                                  self._get_test_executables_grmon)

    def _get_test_executables_grmon(self, _ctx: ItemGetValueContext) -> str:
        return " \\\n".join(
            os.path.basename(executable.path)
            for executable in self._executables)

    def describe(self) -> str:
        return """This test procedure creates an archive with test programs, a
GRMON test script, and a shell script to run the tests manually."""

    def run_tests(self, executables: list[Executable]) -> list[Report]:
        base = self["script-base-path"]
        dir_name = os.path.basename(base)
        grmon_name = f"{base}.grmon"
        shell_name = f"{base}.sh"
        tar_name = f"{base}.tar.xz"
        os.makedirs(os.path.dirname(base), exist_ok=True)
        with tarfile.open(tar_name, "w:xz") as tar_file:
            with open(grmon_name, "w", encoding="utf-8") as grmon_file:
                grmon_file.write(self["grmon-script"])
            tar_file.add(grmon_name, os.path.join(dir_name, "run.grmon"))
            with open(shell_name, "w", encoding="utf-8") as shell_file:
                shell_file.write(self["shell-script"])
            tar_file.add(shell_name, os.path.join(dir_name, "run.sh"))
            for executable in executables:
                tar_file.add(
                    executable.path,
                    os.path.join(dir_name, os.path.basename(executable.path)))
        raise IOError(f"Run the tests provided by {tar_name}")


class _Job:
    # pylint: disable=too-few-public-methods
    def __init__(self, executable: Executable, command: list[str]):
        self.report: Report = {
            "executable": executable.path,
            "executable-sha512": executable.digest,
            "command-line": command
        }
        self.timeout = executable.timeout


def _worker(work_queue: queue.Queue, item: BuildItem):
    discard_patterns = item.item["discard-patterns"]
    while True:
        try:
            job = work_queue.get_nowait()
        except queue.Empty:
            return
        logging.info("%s: run: '%s'", item.uid,
                     "' '".join(job.report["command-line"]))
        job.report["start-time"] = now_utc()
        begin = time.monotonic()
        try:
            process = subprocess_run(job.report["command-line"],
                                     check=False,
                                     stdin=subprocess.DEVNULL,
                                     stdout=subprocess.PIPE,
                                     timeout=job.timeout)
            stdout = process.stdout.decode("latin-1")
        except subprocess.TimeoutExpired as timeout:
            job.report["error"] = "timeout"
            if timeout.stdout is not None:
                stdout = timeout.stdout.decode("latin-1")
            else:
                stdout = ""
        except Exception as err:  # pylint: disable=broad-exception-caught
            job.report["error"] = str(err)
            stdout = ""
        for pattern in discard_patterns:
            if is_enabled([Path(job.report["executable"]).name],
                          pattern["enabled-by"]) and re.search(
                              pattern["pattern"], stdout, re.DOTALL):
                job.report["error"] = ("discarded due to "
                                       f"match with: {pattern['pattern']}")
        output = stdout.rstrip().replace("\r\n", "\n").split("\n")
        job.report["output"] = output
        job.report["duration"] = time.monotonic() - begin
        logging.debug("%s: done: %s", item.uid, job.report["executable"])
        work_queue.task_done()


class SubprocessTestRunner(TestRunner):
    """ Runs tests in subprocesses. """

    def describe(self) -> str:
        self._executable = "${test-program}"
        command = " ".join(f"'{part}'" if " " in part else part
                           for part in self["command"])
        return f"""For each test program (indicated by ``${{test-program}}``),
this test procedure runs the following command as a subprocess on the machine
building the package and captures the output:

.. code-block:: none

    {command}"""

    def run_tests(self, executables: list[Executable]) -> list[Report]:
        work_queue: queue.Queue[_Job] = queue.Queue()
        jobs: list[_Job] = []
        for executable in executables:
            self._executable = executable.path
            job = _Job(executable, self["command"])
            jobs.append(job)
            work_queue.put(job)
        worker_count = int(
            math.ceil(multiprocessing.cpu_count() /
                      float(self["max-process-count"])))
        logging.info("%s: use %s worker threads", self.uid, worker_count)
        for _ in range(min(worker_count, len(executables))):
            threading.Thread(target=_worker,
                             args=(work_queue, self),
                             daemon=True).start()
        work_queue.join()
        return [job.report for job in jobs]

    def get_run_command(self, executable: str) -> None | list[str]:
        self._executable = executable
        return self["command"]



# SPDX-License-Identifier: BSD-2-Clause
""" Links SDD elements to other documents. """

# Copyright (C) 2022 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools

from specitems import Item
from specware import CContent

from .directorystate import DirectoryState
from .linkhub import LinkHub, SpecMapper
from .pkgitems import PackageBuildDirector


def _add_variant(content: CContent, variant: dict) -> None:
    with content.doxygen_block():
        content.add(variant["doxygen"])
        for group in variant.get("groups", []):
            content.add(f"@ingroup {group[group.find('/') + 1:]}")
        links = [
            f"[{item.spec}]({item.view['document-links']['url']})"
            for item in sorted(variant["items"])
        ]
        line = "This design element is related to"
        content_2 = CContent()
        if len(links) > 1:
            line = f"{line}:"
            content_2.add_list(links, line)
        else:
            content_2.add(f"{line} {links[0]}.")
        content.add_paragraph("Traceability", content_2)


class SDDLinker(DirectoryState):
    """ Links SDD elements to other documents. """

    def __init__(self, director: PackageBuildDirector, item: Item) -> None:
        super().__init__(director, item, SpecMapper("sdd", self, item))

    def run(self):
        content = CContent()
        link_hub = self.input("link-hub")
        assert isinstance(link_hub, LinkHub)
        for _, info in sorted(link_hub.name_info.items()):
            if not isinstance(info, dict):
                continue
            for variant in itertools.chain([info], info.get("variants", [])):
                _add_variant(content, variant)
        content.write(self.file)

        self.description.add(f"""Produce the header file
{self.description.path(self.file)} with links from software design elements to
the specification documents.""")



# SPDX-License-Identifier: BSD-2-Clause
""" Collects memory usage benchmarks and object sizes. """

# Copyright (C) 2023 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
from typing import Any, Dict

from specitems import Item, ItemGetValueContext

from .directorystate import DirectoryState
from .membench import gather_sections, gather_object_sizes
from .pkgitems import PackageBuildDirector


class MembenchCollector(DirectoryState):
    """ Collects memory usage benchmarks and object sizes. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self._arch = ""
        self.mapper.add_get_value(f"{self.item.type}:/arch", self._get_arch)

    def _get_arch(self, _ctx: ItemGetValueContext) -> str:
        return self._arch

    def run(self) -> None:
        # Get the memory benchmark results for each build directory
        item_cache = self.item.cache
        results: Dict[str, Dict[str, Any]] = {}
        for link, _ in self.input_links("membench-build"):
            self._arch = self.substitute(link["arch"], link.item)
            path = self.substitute(link["path"], link.item)
            build_label = self.substitute(link["build-label"], link.item)
            logging.info("%s: get memory benchmarks for %s from: %s", self.uid,
                         build_label, path)
            objdump = self["objdump"]
            gdb = self["gdb"]
            assert build_label not in results
            results[build_label] = {
                "membench": gather_sections(item_cache, path, objdump, gdb),
                "object-sizes": gather_object_sizes(item_cache, path, gdb)
            }

        # Save the memory benchmark results
        self.json_dump(results)

        self.description.add(f"""Gather memory benchmark results and produce
the summary file {self.description.path(self.file)}""")



# SPDX-License-Identifier: BSD-2-Clause
""" Builds test report documents. """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from contextlib import contextmanager
import functools
import os
import re
from typing import Any, Callable, Iterator, NamedTuple

from specitems import (EmptyItem, Item, ItemGetValueContext, SphinxContent,
                       base64_to_hex_text)

from .directorystate import DirectoryState
from .docbuilder import DocumentBuilder
from .pkgitems import PackageBuildDirector
from .perfimages import environment_order
from .testaggregator import TestAggregator
from .util import duration

_Failures = dict[str, dict[tuple[Item, str], dict[str, set[str]]]]

_NON_ORDINARY = re.compile(r"[^\x20-\x7e]")


def _escape_char(match: re.Match[str]) -> str:
    return f"\\x{ord(match.group(0)):02x}"


def _invisible_spaces(text: str) -> str:
    return "\u200b".join(iter(_NON_ORDINARY.sub(_escape_char, text)))


def _ok(good: bool) -> str:
    if good:
        return "OK"
    return "NOK"


def _check(expected: str, reported: str) -> str:
    if reported == "?":
        return "NOK"
    return _ok(expected == reported)


def _check_target_hash(expected: str, reported: str) -> str:
    return _ok(reported in ("\u200b", expected))


def _check_gt_zero(_expected: str, reported: str) -> str:
    try:
        if float(reported) > 0.0:
            return "OK"
    except ValueError:
        pass
    return "NOK"


def _check_eq_zero(_expected: str, reported: str) -> str:
    try:
        if int(reported) == 0:
            return "OK"
    except ValueError:
        pass
    return "NOK"


def _check_duration(_expected: str, reported: str) -> str:
    if reported == "?":
        return "NOK"
    return "OK"


def _listed(listed: bool) -> str:
    if listed:
        return "listed"
    return "not listed"


def _zero_one(value: bool) -> str:
    if value:
        return "1"
    return "0"


def _rtems_version_to_commit(version: str) -> str:
    return _invisible_spaces(version.split(".")[-1])


_RSB_COMMIT = re.compile(r"RSB ([^,]+),")


def _rsb_version_to_commit(version: str) -> str:
    match = _RSB_COMMIT.search(version)
    if match is None:
        return "?"
    return _invisible_spaces(match.group(1))


def _option(option: str, options: list[str]) -> str:
    return _listed(option in options)


def _target_hash(target_hash: str) -> str:
    if not target_hash:
        return "\u200b"
    return _invisible_spaces(target_hash)


_PROPERTY_TRANSFORM = {
    "rtems-source-builder-version": _rsb_version_to_commit,
    "rtems-version": _rtems_version_to_commit
}

_HEADER = ["Property", "Line", "Reported", "Expected", "Status"]

_WIDTHS = [42, 8, 20, 20, 10]

_ERRORS = {
    "no-begin-of-test-message":
    "The test output contains no begin of test message.",
    "no-end-of-test-message":
    "The test output contains no end of test message.",
    "unexpected-bsp":
    "The BSP has not the expected name.",
    "unexpected-build":
    "At least one build configuration option has not the expected value "
    "or the build configuration information is not present.",
    "unexpected-build-label":
    "The build label has not the expected value.",
    "unexpected-compiler":
    "The compiler version has not the expected value.",
    "unexpected-duration":
    "The test duration has not the expected value.",
    "unexpected-failed-steps-count":
    "The failed test steps count value is not zero.",
    "unexpected-report-hash":
    "The report hash has not the expected value.",
    "unexpected-rtems-debug":
    "The RTEMS_DEBUG build configuration option has not the expected value.",
    "unexpected-rtems-multiprocessing":
    "The RTEMS_MULTIPROCESSING build configuration option "
    "has not the expected value.",
    "unexpected-rtems-posix-api":
    "The RTEMS_POSIX_API build configuration option "
    "has not the expected value.",
    "unexpected-rtems-profiling":
    "The RTEMS_PROFILING build configuration option "
    "has not the expected value.",
    "unexpected-rtems-smp":
    "The RTEMS_SMP build configuration option has not the expected value.",
    "unexpected-runtime-maximum":
    "A maximum runtime value is greater than expected.",
    "unexpected-runtime-median":
    "A median runtime value is not in the expected interval.",
    "unexpected-runtime-minimum":
    "A minimum runtime value is less than expected.",
    "unexpected-step-count":
    "The test step count value is not positive.",
    "unexpected-target-hash":
    "The target hash has not the expected value.",
    "unexpected-tools":
    "The tools version has not the expected value.",
    "unexpected-version":
    "The RTEMS Git commit has not the expected value.",
}


class _TestProperty(NamedTuple):
    name: str
    type: str
    expected: str


class _TestContext:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, reporter: "TestReporter"):
        self.content = SphinxContent()
        self.enabled_set = reporter.enabled_set
        self.expected_failures: _Failures = {}
        self.unexpected_failures: _Failures = {}
        self.limits_uid = ""
        self.limits_by_req: dict[str, dict] = {}
        self.target_hash = ""
        self.target_uid = ""
        self.build_label = ""
        self.bsp: str = reporter.component["bsp"]
        self.properties: dict[str, _TestProperty] = {}
        for link, _ in reporter.input_links("test-property"):
            data = reporter.substitute(link.data)
            self.properties[data["key"]] = _TestProperty(
                data["property-name"], data["type"],
                _invisible_spaces(data["expected"]))
        self.item = EmptyItem()
        self.verifications: dict[str, str] = {}
        self.output_label = ""
        self.config_data: dict[str, Any] = {}

    def begin_report(self, report: dict) -> None:
        """ Begins the report. """
        self.output_label = f"{self.content.get_label()}Output"
        executable = os.path.basename(report["executable"])
        self.content.add(f"""This report was produced by the
:file:`{executable}`
executable.  The executable file had an SHA512 digest of
{base64_to_hex_text(report['executable-sha512'])}.""")
        error = report.get("error")
        if error:
            with self.content.directive("error", value="Test runner error"):
                self.content.add(error)
        else:
            self.content.gap = False

    def output_line_ref(self, line: int | str) -> str:
        """ Gets the output line reference for the line. """
        if isinstance(line, str):
            return line
        return (f":ref:`{line + 1} <"
                f"{self.output_label}{line - line % 100}>`")

    def add_error(self, error: str, add_to_content: bool = False) -> None:
        """
        Add the error with the reason.

        Optionally, add the reason to the content.
        """
        text = _ERRORS[error]
        if add_to_content:
            self.content.wrap(text)
        key = self.config_data["config-key"]
        config = f":ref:`Configuration - {key} <{self.content.get_label()}>`"
        uid = self.verifications.get(self.item.uid, None)
        failures = self.unexpected_failures
        verification_text = ""
        if uid is not None:
            verification = self.item.cache[uid]
            if error in verification["acceptable-test-errors"]:
                failures = self.expected_failures
                verification_text = verification["text"]
        failures.setdefault(self.target_uid, {}).setdefault(
            (self.item, verification_text), {}).setdefault(config,
                                                           set()).add(text)

    def add_failures(self, which: str, failures: _Failures,
                     test_aggregator: TestAggregator) -> None:
        """ Add the test failures to the content. """
        with self.content.section(f"List of {which} test failures"):
            if not failures:
                self.content.add(f"There were no {which} test errors "
                                 "found in the test outputs.")
                return
            for target_uid, by_test in failures.items():
                target_data = test_aggregator.targets[target_uid]
                target_section = f"Target - {target_data['name']}"
                with self.content.section(target_section):
                    for item_text, by_config in sorted(by_test.items()):
                        with self.content.section(item_text[0].spec_2):
                            self.content.add(item_text[1])
                            for config, errors in sorted(by_config.items()):
                                self.content.add_list_item(f"{config}:")
                                self.content.add_blank_line()
                                with self.content.indent("  "):
                                    self.content.add_list(sorted(errors))

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def check(self,
              rows: list[list[str]],
              info: dict,
              name: str,
              key: str,
              expected: str,
              transform: Callable[[Any], str],
              check=_check) -> None:
        """ Check the test property. """
        try:
            line = self.output_line_ref(info[f"line-{key}"])
        except KeyError:
            line = "?"
            reported = "?"
        else:
            reported = transform(info[key])
        status = check(expected, reported)
        if status != "OK":
            self.add_error(f"unexpected-{key}")
        rows.append([name, line, reported, expected, status])

    def check_property(self, rows: list[list[str]], info: dict, key: str,
                       property_key: str) -> None:
        """ Check the property. """
        prop = self.properties[property_key]
        self.check(rows, info, prop.name, key, prop.expected,
                   _PROPERTY_TRANSFORM.get(prop.type, _invisible_spaces))

    def check_rtems_commit(self, rows: list[list[str]], info: dict) -> None:
        """ Check the RTEMS Git commit. """
        self.check_property(rows, info, "version", "rtems-commit")

    def check_compiler(self, rows: list[list[str]], info: dict,
                       key: str) -> None:
        """ Check the compiler version. """
        self.check_property(rows, info, key, "compiler-version")

    def _check_build_info(self, rows: list[list[str]], info: dict) -> None:
        for option in [
                "RTEMS_DEBUG", "RTEMS_MULTIPROCESSING", "RTEMS_PARAVIRT",
                "RTEMS_POSIX_API", "RTEMS_PROFILING", "RTEMS_SMP"
        ]:
            self.check(rows, info, option, "build",
                       _listed(option in self.enabled_set),
                       functools.partial(_option, option))

    def check_build_options(self, rows: list[list[str]], info: dict) -> None:
        """ Check the test suite build options. """
        for option in [
                "debug", "multiprocessing", "posix-api", "profiling", "smp"
        ]:
            key = f"rtems-{option}"
            name = key.replace("-", "_").upper()
            self.check(rows, info, name, key,
                       _zero_one(name in self.enabled_set), _zero_one)

    def check_validations_by_test(self, test_aggregator: TestAggregator,
                                  target_uid: str) -> None:
        """
        Check that each validation by test has a test result for the target.
        """
        for item in test_aggregator.spec.related_validations_by_test:
            if target_uid not in item.view.get("test-results", {}):
                text = "There are no test results available for this target."
                verification_uid = self.verifications.get(item.uid, None)
                failures = self.unexpected_failures
                if verification_uid is not None:
                    verification = self.item.cache[verification_uid]
                    if "no-test-results" in verification[
                            "acceptable-test-errors"]:
                        failures = self.expected_failures
                        text = verification["text"]
                failures.setdefault(self.target_uid, {}).setdefault(
                    (item, text), {})

    def add_table(self, rows: list[list[str]], widths: list[int]) -> None:
        """ Add a table to the content with the rows and widths. """
        self.content.add_grid_table(rows, widths, font_size=-3)

    def add_test_info(self, info: dict) -> None:
        """ Add the test information to the content. """
        rows = [_HEADER]
        try:
            begin = self.output_line_ref(info["line-begin-of-test"])
            self.content.wrap(
                f"There is a valid begin of test message at line {begin}.")
        except KeyError:
            self.add_error("no-begin-of-test-message", True)
        self.content.gap = False
        try:
            end = self.output_line_ref(info["line-end-of-test"])
            self.content.wrap(f"""There is a valid end of test message at line
{end}.  This indicates that the test program executed without a detected
error.""")
        except KeyError:
            self.add_error("no-end-of-test-message", True)
        self.content.gap = False
        self.content.wrap("""The following table lists an evaluation of the
reported test information.""")
        self.check_rtems_commit(rows, info)
        self._check_build_info(rows, info)
        self.check_compiler(rows, info, "tools")
        self.add_table(rows, _WIDTHS)

    def add_limits(self, env_name: str, env_data: dict, limits: dict) -> None:
        """ Add the runtime performance limits to the content. """
        rows = [["Limit Kind", "Specified Limits", "Actual Value", "Status"]]
        value = env_data["min"]
        lower_bound = limits[env_name]["min-lower-bound"]
        status = _ok(lower_bound <= value)
        if status != "OK":
            self.add_error("unexpected-runtime-minimum")
        rows.append([
            "Minimum", f"{duration(lower_bound)} :math:`\\leq` Minimum",
            duration(value), status
        ])
        value = env_data["q2"]
        lower_bound = limits[env_name]["median-lower-bound"]
        upper_bound = limits[env_name]["median-upper-bound"]
        status = _ok(lower_bound <= value <= upper_bound)
        if status != "OK":
            self.add_error("unexpected-runtime-median")
        rows.append([
            "Median", f"{duration(lower_bound)} :math:`\\leq` Median "
            f":math:`\\leq` {duration(upper_bound)}",
            duration(value), status
        ])
        value = env_data["max"]
        upper_bound = limits[env_name]["max-upper-bound"]
        status = _ok(value <= upper_bound)
        if status != "OK":
            self.add_error("unexpected-runtime-maximum")
        rows.append([
            "Maximum", f"Maximum :math:`\\leq` {duration(upper_bound)}",
            duration(value), status
        ])
        self.add_table(rows, [15, 45, 25, 15])

    @contextmanager
    def file_scope(self, file_path: str, data: dict) -> Iterator[str]:
        """ Opens a file scope. """
        file_name = os.path.basename(data["report-file"])
        self.content.add(file_name)
        content = self.content
        self.content = SphinxContent()
        yield file_name
        self.content.write(
            os.path.join(os.path.dirname(file_path), f"{file_name}.rst"))
        self.content = content


def _add_output(ctx: _TestContext, report: dict) -> None:
    ctx.content.add_program_output(report["output"], report["data-ranges"],
                                   ctx.output_label)


def _add_test_output(ctx: _TestContext, report: dict) -> None:
    with ctx.content.section("Test output"):
        ctx.content.add(
            "The test report was generated from the following test output.")
        _add_output(ctx, report)


def _save_unexpected_failures(destination: DirectoryState,
                              failures: _Failures) -> None:
    target_to_failures: dict[str, list[str]] = {}
    for target_uid, by_test in sorted(failures.items()):
        target_to_failures[target_uid] = [
            item_text[0].uid for item_text in sorted(by_test.keys())
        ]
    destination["unexpected-test-failures"] = target_to_failures


class TestReporter(DocumentBuilder):
    """ Builds a test report. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.mapper.add_get_value(f"{self.item.type}:/reports", self._reports)

    def _add_runtime_measurements(self, ctx: _TestContext,
                                  suite_data: dict) -> dict[str, str]:
        uid_to_label: dict[str, str] = {}
        for req_uid, measurement_data in sorted(
                suite_data["runtime-measurements"].items()):
            req = self.item.cache[req_uid]
            with ctx.content.section(f"Runtime measurement - {req.spec_2}",
                                     label=measurement_data["label"]):
                ctx.content.wrap(f"""For the runtime performance requirement
{self.mapper.get_link(req, 'test-plan')}, the following runtime values were
measured on this target and configuration in the listed measurement
environments.""")
                ctx.content.add_image(
                    os.path.relpath(f"{measurement_data['boxplot']}.*",
                                    os.path.dirname(self.file_path)), "50%")
                uid_to_label[req_uid] = ctx.content.get_label()
                limits = ctx.limits_by_req[req.uid]
                for env_name, env_data in sorted(
                        measurement_data["variants"].items(),
                        key=environment_order):
                    ctx.content.add_label(env_data["label"])
                    ctx.content.add_rubric(
                        f"Measurement environment - {env_name}")
                    begin = ctx.output_line_ref(env_data["line-begin"])
                    end = ctx.output_line_ref(env_data["line-end"])
                    ctx.content.add(f"""The runtime measurement report for this
measurement environment was generated from lines {begin} up to and including
{end} of the test output.""")
                    ctx.content.add_image(
                        os.path.relpath(f"{env_data['histogram']}.*",
                                        os.path.dirname(self.file_path)),
                        "50%")
                    ctx.add_limits(env_name, env_data, limits)
        return uid_to_label

    def _add_remarks(self, ctx: _TestContext, remarks: list) -> None:
        if not remarks:
            return
        test_cases: list[tuple[str, str]] = []
        for remark in remarks:
            try:
                uid = remark["uid"]
            except KeyError:
                pass
            else:
                link = self.mapper.get_link(self.item.cache[uid], "test-plan")
                line = ctx.output_line_ref(remark["line"])
                test_cases.append((link, line))
        if len(test_cases) == 1:
            test_case = test_cases[0]
            ctx.content.append(f"""It runs the parameterized test case
{test_case[0]} reported in line {test_case[1]}.""")
            return
        ctx.content.append("It runs the following parameterized test cases:")
        for test_case in test_cases:
            ctx.content.add_list_item(
                f"{test_case[0]} reported in line {test_case[1]}")

    def _add_test_cases(self, ctx: _TestContext, suite_data: dict,
                        uid_to_label: dict[str, str]) -> None:
        for case_uid, case_data in sorted(suite_data["test-cases"].items()):
            case_item = self.item.cache[case_uid]
            ctx.item = case_item
            with ctx.content.section(f"Test case - {case_item.spec_2}",
                                     label=case_data["label"]):
                ctx.content.add(f"""This test case is specified by
{self.mapper.get_link(case_item, 'test-plan')}.""")
                self._add_remarks(ctx, case_data["remarks"])
                begin = ctx.output_line_ref(case_data["line-begin"])
                end = ctx.output_line_ref(case_data["line-end"])
                ctx.content.add(f"""The following table lists an evaluation of
the test case information reported in lines {begin} up to and including {end}
of the test output.""")
                rows = [_HEADER]
                ctx.check(rows, case_data, "Step Count", "step-count", "> 0",
                          str, _check_gt_zero)
                ctx.check(rows, case_data, "Failed Steps Count",
                          "failed-steps-count", "0", str, _check_eq_zero)
                ctx.check(rows, case_data, "Duration", "duration",
                          ":math:`\\geq` 0", duration, _check_duration)
                ctx.add_table(rows, _WIDTHS)
                uids = set(
                    measurement_data["requirement-uid"]
                    for measurement_data in case_data["runtime-measurements"])
                if uids:
                    ctx.content.wrap("""This test case contains the
following runtime measurements presented in the preceeding sections:""")
                    for uid in sorted(uids):
                        item = self.item.cache[uid]
                        ctx.content.add_list_item(
                            f":ref:`{item.spec_2} <{uid_to_label[uid]}>`")

    def _add_one_test_suite(self, ctx: _TestContext, suite_uid: str,
                            suite_data: dict) -> None:
        suite_item = self.item.cache[suite_uid]
        ctx.item = suite_item
        suite_section = f"Test suite - {suite_item.spec_2}"
        with ctx.content.section(suite_section, label=suite_data["label"]):
            report = suite_data["report"]
            ctx.begin_report(report)
            ctx.content.wrap(f"""This test suite is specified by
{self.mapper.get_link(suite_item)}.""")
            ctx.content.gap = False
            ctx.add_test_info(report["info"])
            begin = ctx.output_line_ref(report["test-suite"]["line-begin"])
            end = ctx.output_line_ref(report["test-suite"]["line-end"])
            ctx.content.wrap(f"""The following table lists an evaluation of the
test suite information reported in lines {begin} up to and including {end} of
the test output.""")
            if suite_data["runtime-measurements"]:
                ctx.content.gap = False
                ctx.content.wrap("""The runtime measurements and test cases
of this test suite are presented in the following sections.""")
            else:
                ctx.content.gap = False
                ctx.content.wrap("""The test cases of this test suite are
presented in the following sections.""")
            rows = [_HEADER]
            info = report["test-suite"]
            ctx.check_compiler(rows, info, "compiler")
            ctx.check_rtems_commit(rows, info)
            ctx.check(rows, info, "BSP", "bsp", _invisible_spaces(ctx.bsp),
                      _invisible_spaces)
            ctx.check(rows, info, "Build Label", "build-label",
                      ctx.build_label, _invisible_spaces)
            ctx.check(rows, info, "Target Hash", "target-hash",
                      ctx.target_hash, _target_hash, _check_target_hash)
            ctx.check_build_options(rows, info)
            ctx.check(rows, info, "Step Count", "step-count", "> 0", str,
                      _check_gt_zero)
            ctx.check(rows, info, "Failed Steps Count", "failed-steps-count",
                      "0", str, _check_eq_zero)
            ctx.check(rows, info, "Duration", "duration", ":math:`\\geq` 0",
                      duration, _check_duration)
            ctx.check(rows, info, "Report Hash", "report-hash",
                      _invisible_spaces(info["report-hash-calculated"]),
                      _invisible_spaces)
            ctx.add_table(rows, _WIDTHS)
            uid_to_label = self._add_runtime_measurements(ctx, suite_data)
            self._add_test_cases(ctx, suite_data, uid_to_label)
            _add_test_output(ctx, report)

    def _add_test_suites(self, ctx: _TestContext) -> None:
        for suite_uid, suite_data in sorted(
                ctx.config_data["test-suites"].items()):
            with ctx.file_scope(self.file_path, suite_data):
                self._add_one_test_suite(ctx, suite_uid, suite_data)

    def _add_test_programs(self, ctx: _TestContext) -> None:
        for uid, report in sorted(ctx.config_data["test-programs"].items()):
            with ctx.file_scope(self.file_path, report) as file_name:
                program_item = self.item.cache[uid]
                ctx.item = program_item
                program_section = f"Test program - {program_item.spec_2}"
                with ctx.content.section(program_section, label=file_name):
                    ctx.begin_report(report)
                    ctx.add_test_info(report["info"])
                    for image in report.get("images", []):
                        ctx.content.add_image(
                            os.path.relpath(f"{image}.*",
                                            os.path.dirname(self.file_path)),
                            "50%")
                    _add_test_output(ctx, report)

    def _add_other_programs(self, ctx: _TestContext) -> None:
        for executable, report in sorted(
                ctx.config_data["other-programs"].items()):
            with ctx.file_scope(self.file_path, report) as file_name:
                program_section = f"Other program - {executable}"
                with ctx.content.section(program_section, label=file_name):
                    ctx.begin_report(report)
                    _add_output(ctx, report)

    def _add_coverage(self, ctx: _TestContext,
                      test_aggregator: TestAggregator) -> None:
        with ctx.content.section("Coverage data"):
            coverage_count = 0
            with ctx.content.directive("toctree"):
                report = {"report-file": "coverage"}
                with ctx.file_scope(self.file_path, report):
                    with ctx.content.label_scope("Coverage"):
                        for target_data in test_aggregator.targets.values():
                            ctx.verifications = target_data[
                                "test-error-verifications"]
                            target_uid = target_data["uid"]
                            target_item = self.item.cache[target_uid]
                            target_section = f"Target - {target_data['name']}"
                            failure_key = (
                                target_item,
                                "For this target, the following code "
                                "coverage issues were present.")
                            issues: dict[
                                str, set[str]] = ctx.unexpected_failures.get(
                                    target_uid, {}).get(failure_key, {})
                            with ctx.content.section(target_section):
                                for config_data in target_data["configs"]:
                                    if "coverage" not in config_data:
                                        continue
                                    coverage_count += 1
                                    test_aggregator.add_coverage_of_config(
                                        ctx.content, config_data, issues)
                            if issues:
                                ctx.unexpected_failures.setdefault(
                                    target_uid, {})[failure_key] = issues
            if coverage_count == 0:
                ctx.content.add("There is no coverage data available.")

    def _reports(self, _unused: ItemGetValueContext) -> str:
        test_aggregator = self.input("test-aggregation")
        assert isinstance(test_aggregator, TestAggregator)
        ctx = _TestContext(self)
        for target_uid, target_data in test_aggregator.targets.items():
            ctx.limits_uid = target_data["limits-uid"]
            ctx.limits_by_req = target_data["limits-by-requirement"]
            ctx.target_hash = _invisible_spaces(target_data["target-hash"])
            ctx.target_uid = target_uid
            ctx.verifications = target_data["test-error-verifications"]
            with ctx.content.section(f"Target - {target_data['name']}",
                                     label=target_data["label"]):
                with ctx.content.section("Test procedure description"):
                    ctx.content.add(target_data["test-runner-description"])
                for config_data in target_data["configs"]:
                    ctx.config_data = config_data
                    config_key = config_data["config-key"]
                    config_section = f"Configuration - {config_key}"
                    ctx.build_label = _invisible_spaces(
                        config_data["build-label"])
                    with ctx.content.section(config_section,
                                             label=config_data["label"]):
                        with ctx.content.directive("toctree"):
                            self._add_test_suites(ctx)
                            self._add_test_programs(ctx)
                            self._add_other_programs(ctx)
            ctx.check_validations_by_test(test_aggregator, target_uid)
        self._add_coverage(ctx, test_aggregator)
        ctx.add_failures("expected", ctx.expected_failures, test_aggregator)
        ctx.add_failures("unexpected", ctx.unexpected_failures,
                         test_aggregator)
        _save_unexpected_failures(self, ctx.unexpected_failures)
        return ctx.content.join()



# SPDX-License-Identifier: BSD-2-Clause
""" This is the specware package. """

# Copyright (C) 2019, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .archiver import *  # noqa: F401, F403
from .dirstatebase import *  # noqa: F401, F403
from .directorystate import *  # noqa: F401, F403
from .doxyfile import *  # noqa: F401, F403
from .gcdaproducer import *  # noqa: F401, F403
from .icdbuilder import *  # noqa: F401, F403
from .itemcachestate import *  # noqa: F401, F403
from .linkhub import *  # noqa: F401, F403
from .membench import *  # noqa: F401, F403
from .membenchcollector import *  # noqa: F401, F403
from .packagechanges import *  # noqa: F401, F403
from .packagemanual import *  # noqa: F401, F403
from .perfimages import *  # noqa: F401, F403
from .pkgitems import *  # noqa: F401, F403
from .pkgfactory import *  # noqa: F401, F403
from .pkgtemplate import *  # noqa: F401, F403
from .pkgworkspace import *  # noqa: F401, F403
from .reposubset import *  # noqa: F401, F403
from .rtems import *  # noqa: F401, F403
from .rtemstestsimages import *  # noqa: F401, F403
from .runactions import *  # noqa: F401, F403
from .runexecutablecmd import *  # noqa: F401, F403
from .sddlinker import *  # noqa: F401, F403
from .sourcecompare import *  # noqa: F401, F403
from .sourcetospec import *  # noqa: F401, F403
from .spamrmanager import *  # noqa: F401, F403
from .speccompare import *  # noqa: F401, F403
from .sphinxbuilder import *  # noqa: F401, F403
from .sreldbuilder import *  # noqa: F401, F403
from .srsbuilder import *  # noqa: F401, F403
from .testaggregator import *  # noqa: F401, F403
from .testoutputparser import *  # noqa: F401, F403
from .testreporter import *  # noqa: F401, F403
from .testrunner import *  # noqa: F401, F403
from .testrunneresa import *  # noqa: F401, F403
from .testrunnergrpc import *  # noqa: F401, F403
from .util import *  # noqa: F401, F403



# SPDX-License-Identifier: BSD-2-Clause
""" Builds a Software Release Document (SRelD). """

# Copyright (C) 2023 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from specitems import Item, ItemGetValueContext

from .docbuilder import DocumentBuilder
from .pkgitems import PackageBuildDirector
from .packagechanges import PackageChanges


class SRelDBuilder(DocumentBuilder):
    """ Builds a Software Release Document (SRelD). """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/changes", self._get_changes)
        self.mapper.add_get_value(f"{my_type}:/issues", self._get_issues)

    def _get_changes(self, _ctx: ItemGetValueContext) -> str:
        changes = self.input("package-changes")
        assert isinstance(changes, PackageChanges)
        return changes.get_current_changes(self.mapper)

    def _get_issues(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            changes = self.input("package-changes")
            assert isinstance(changes, PackageChanges)
            return changes.get_current_issues(self.mapper, self.section_level)



# SPDX-License-Identifier: BSD-2-Clause
""" This module provides a test runner using gRPC.  """

# Copyright (C) 2024, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
from pathlib import Path
import subprocess
import tempfile
import time
import grpc  # type: ignore

from specitems import is_enabled, Item, ItemGetValueContext

# pylint: disable=no-name-in-module
from spectestrunner import (GRPCDescribeTargetRequest, GRPCRunImageRequest,
                            GRPCServiceStub)  # type: ignore

from .pkgitems import PackageBuildDirector
from .testrunner import Executable, Report, TestRunner
from .util import now_utc


def _get_symbols(exe_path: str, nm_path: str) -> dict[str, list[int]]:
    """ Return the symbols of the executable using the nm tool. """
    result = subprocess.run([nm_path, exe_path],
                            check=True,
                            capture_output=True,
                            text=True)
    symbols: dict[str, list[int]] = {}
    for line in result.stdout.split("\n"):
        try:
            address, _, name = line.split(None, 2)
        except ValueError:
            pass
        else:
            try:
                symbols.setdefault(name, []).append(int(address, 16))
            except ValueError:
                pass
    return symbols


class GRPCTestRunner(TestRunner):
    """ Runs the tests using an agent through gRPC.  """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.mapper.add_get_value(f"{self.item.type}:/target-name",
                                  self._get_target_name)

    def _get_optional_value(self, key: str) -> str:
        for optional_value in self[key]:
            if is_enabled(self.enabled_set, optional_value["enabled-by"]):
                return optional_value["value"]
        raise ValueError(f"{self.uid}: there is no value for attribute {key} "
                         f"and enabled set {sorted(self.enabled_set)}")

    def _get_target_name(self, _ctx: ItemGetValueContext) -> str:
        return self._get_optional_value("target")

    def describe(self) -> str:
        server_address = self._get_optional_value("server-address")
        target_id = self._get_optional_value("target")
        with grpc.insecure_channel(server_address) as channel:
            stub = GRPCServiceStub(channel)
            try:
                response = stub.request_describe_target(
                    GRPCDescribeTargetRequest(target_id=target_id))
                logging.debug("%s: received target description for: %s",
                              self.uid, target_id)
            except grpc.RpcError as err:
                return (
                    f"Requesting the target description of ``{target_id}`` "
                    f"from the test server at {server_address} "
                    f"failed: {err}.")
            return response.description

    def _get_executable_data(self, executable: Executable) -> bytes:
        original = Path(executable.path)
        strip = self["strip"]
        if strip is None:
            return original.read_bytes()
        with tempfile.TemporaryDirectory() as directory:
            stripped = Path(directory) / "stripped"
            subprocess.run(
                [self["strip"], "-g", "-o",
                 str(stripped),
                 str(original)],
                check=True)
            return stripped.read_bytes()

    # pylint: disable=too-many-locals
    def run_tests(self, executables: list[Executable]) -> list[Report]:
        reports: list[Report] = []
        target_id = self._get_optional_value("target")
        nm = self["nm"]
        reset = self._get_optional_value("reset")
        server_address = self._get_optional_value("server-address")
        with grpc.insecure_channel(server_address) as channel:
            stub = GRPCServiceStub(channel)
            for executable in executables:
                start_time = now_utc()
                monotonic_begin = time.monotonic()
                logging.debug("%s: send request for: %s", self.uid,
                              executable.path)
                breakpoints = _get_symbols(executable.path, nm)[reset]
                report: Report = {
                    "command-line": [],
                    "executable": executable.path,
                    "executable-sha512": executable.digest,
                    "start-time": start_time
                }
                try:
                    response = stub.request_run_image(
                        GRPCRunImageRequest(
                            target_id=target_id,
                            breakpoints=breakpoints,
                            path=executable.path,
                            digest=executable.digest,
                            data=self._get_executable_data(executable),
                            execution_timeout_in_seconds=executable.timeout))
                    logging.debug("%s: received response for: %s", self.uid,
                                  executable.path)
                except grpc.RpcError as err:
                    report["output"] = []
                    report["error"] = str(err)
                else:
                    report["output"] = [
                        line.removesuffix("\r") for line in
                        response.output.decode("latin-1").split("\n")
                    ]
                    if response.status != "success":
                        report["error"] = response.status
                    report["load-duration-in-seconds"] = \
                        response.load_duration_in_seconds
                    report["execution-duration-in-seconds"] = \
                        response.execution_duration_in_seconds
                report["duration"] = time.monotonic() - monotonic_begin
                reports.append(report)
        return reports



# SPDX-License-Identifier: BSD-2-Clause
""" Builds an Interface Control Document (ICD). """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools

from specitems import Item, ItemGetValueContext, SphinxContent

from .pkgitems import PackageBuildDirector
from .specdocbuilder import SpecDocumentBuilder


def _visit_domain(item: Item, interfaces: list[Item]) -> None:
    interfaces.append(item)
    for item_2 in itertools.chain(item.children("interface-placement"),
                                  item.parents("interface-enumerator")):
        _visit_domain(item_2, interfaces)


class ICDBuilder(SpecDocumentBuilder):
    """ Builds an Interface Control Document (ICD). """

    def __init__(self, director: PackageBuildDirector, item: Item) -> None:
        super().__init__(director, item)
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/icd-requirements-and-design",
                                  self._get_requirements_and_design)

    def get_items_of_document(self) -> list[Item]:
        return self.spec.get_related_interfaces()

    def _add_interface_requirements(self, content: SphinxContent) -> None:
        types = ("requirement/non-functional/interface-requirement", )
        for item in self.spec.get_related_items_by_type(types):
            self.add_item(content, item)

    def _add_interface_design(self, content: SphinxContent) -> None:
        types = ("interface/domain", )
        for domain in self.spec.get_related_items_by_type(types):
            with content.section(domain["name"]):
                content.add(self.mapper.substitute(domain["description"]))
                interfaces: list[Item] = []
                _visit_domain(domain, interfaces)
                for item in sorted(interfaces):
                    self.add_item(content, item)

    def _get_requirements_and_design(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level,
                                    the_license=self.content_license)
            with content.section("Requirements and design"):
                with content.section(
                        "General provisions to the requirements in the IRD"):
                    content.add(
                        "There are no general provisions to requirements "
                        "in the :term:`IRD`.")
                with content.section("Interface requirements"):
                    self._add_interface_requirements(content)
                with content.section("Interface design"):
                    self._add_interface_design(content)
            return content.join()



# SPDX-License-Identifier: BSD-2-Clause
""" Builds an archive file from directory states. """

# Copyright (C) 2021 EDISOFT
# Copyright (C) 2020, 2024 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import graphlib
import itertools
import pickle
import logging
import lzma
import os
import stat
import tarfile

from specitems import save_data

from .directorystate import DirectoryState

_FileInfo = list[tuple[str, str, bool]]
_ArchiveFiles = dict[str, _FileInfo]


def _check_for_duplicates(uid: str, file_path: str,
                          file_info: _FileInfo) -> None:
    if len(file_info) == 1:
        return
    logging.info("%s: duplicates in directory states %s for file: %s", uid,
                 list(info[1] for info in file_info), file_path)
    for index, info in enumerate(file_info):
        for info_2 in file_info[index + 1:]:
            if info[0] != info_2[0]:
                logging.error(
                    "%s: inconsistent file hashes %s (of %s) "
                    "and %s (of %s) for file: %s", uid, info[0], info[1],
                    info_2[0], info_2[1], file_path)


_SCRIPT_HEAD = f"""#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
\"\"\" Verifies the files of the package. \"\"\"

# Copyright (C) 2021, 2022 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import binascii
import argparse
import hashlib
import pickle
import logging
import lzma
import os
import sys


# This is a list of (<file-path>, <binary-hash>) tuples,
# which was serialized using the Python pickle protocol \
{pickle.DEFAULT_PROTOCOL},
# then compressed using LZMA,
# and finally encoded as base85.
_B85_DATA = """

_SCRIPT_TAIL = """

def _hash_file(path: str) -> bytes:
    file_hash = hashlib.sha512()
    if os.path.islink(path):
        file_hash.update(os.readlink(path).encode("utf-8"))
    else:
        buf = bytearray(65536)
        memview = memoryview(buf)
        with open(path, "rb", buffering=0) as src:
            for size in iter(lambda: src.readinto(memview), 0):  # type: ignore
                file_hash.update(memview[:size])
    return file_hash.digest()


def _hex(digest: bytes) -> str:
    return binascii.hexlify(digest).decode("ascii")


def _check_file(file_path: str, expected_files: dict[str, bytes]) -> int:
    expected_hash = expected_files[file_path]
    actual_hash = _hash_file(file_path)
    if expected_hash != actual_hash:
        logging.error(
            "expected hash is %s, actual hash is %s for file: %s",
            _hex(expected_hash), _hex(actual_hash), file_path)
        return 1
    return 0


def _verify_files(script: str, expected_files: dict[str, bytes]) -> int:
    status = 0
    script = os.path.normpath(script)
    for path, dirs, files in os.walk("."):
        dirs.sort()
        for name in sorted(files):
            file_path = os.path.normpath(os.path.join(path, name))
            if file_path in expected_files:
                status = _check_file(file_path, expected_files)
                del expected_files[file_path]
            elif file_path != script:
                logging.warning("unexpected file: %s", file_path)
    for maybe_missing in expected_files.keys():
        if os.path.islink(maybe_missing):
            status = _check_file(maybe_missing, expected_files)
            continue
        logging.error("missing file: %s", maybe_missing)
        status = 1
    return status


def main(script: str, argv: list[str]) -> int:
    \"\"\" Verifies the files of the package. \"\"\"
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        type=str.upper,
        default="WARNING",
        help="log level")
    parser.add_argument("--log-file",
                        type=str,
                        default=None,
                        help="log to this file")
    parser.add_argument("--list-files",
                        action="store_true",
                        help="list the files of the package")
    parser.add_argument("--list-files-and-hashes",
                        action="store_true",
                        help="list the files of the package "
                        "with the SHA512 digest of each file")
    args = parser.parse_args(argv)
    logging.basicConfig(filename=args.log_file, level=args.log_level)
    data = pickle.loads(lzma.decompress(base64.b85decode(_B85_DATA)))
    expected_files = dict((os.path.normpath(x[0]), x[1]) for x in data)
    status = 0
    if args.list_files_and_hashes:
        for file_path, hash_value in expected_files.items():
            print(f"{file_path}\t{_hex(hash_value)}")
    elif args.list_files:
        for file_path in expected_files.keys():
            print(file_path)
    else:
        status = _verify_files(script, expected_files)
    return status



if __name__ == "__main__":
    status = main(sys.argv[0], sys.argv[1:])
    sys.exit(status)
"""


def _create_verification_script(script_file: str,
                                archive_files: _ArchiveFiles) -> None:
    script_dir = os.path.dirname(script_file)
    with open(script_file, "w", encoding="utf-8") as out:
        out.write(_SCRIPT_HEAD)
        data = [(os.path.relpath(file_path, script_dir),
                 base64.urlsafe_b64decode(info[-1][0]))
                for file_path, info in sorted(archive_files.items())]
        pickle_data = pickle.dumps(data)
        lzma_data = lzma.compress(pickle_data)
        b85_data = base64.b85encode(lzma_data).decode("ascii")
        for i in range(0, len(b85_data), 76):
            out.write(f"\\\n\"{b85_data[i:i + 76]}\"")
        out.write(_SCRIPT_TAIL)
    os.chmod(script_file, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)


class Archiver(DirectoryState):
    """
    The archiver adds the file of its directory state member inputs to an
    archive file.
    """

    def _gather_files(self, dir_state: DirectoryState,
                      dependencies: dict[str, set[str]],
                      archive_files: _ArchiveFiles) -> None:
        dir_state_uid = dir_state.uid
        logging.info("%s: gather files of directory state: %s", self.uid,
                     dir_state_uid)
        for file_path, hash_value in dir_state.files_and_hashes():
            link_targets = dependencies.setdefault(file_path, set())
            try:
                target = os.readlink(file_path)
            except OSError:
                pass
            else:
                link_targets.add(
                    os.path.normpath(
                        os.path.join(os.path.dirname(file_path), target)))
            assert hash_value
            archive_files.setdefault(file_path, []).append(
                (hash_value, dir_state_uid, os.path.islink(file_path)))

    def _export_spec(self, dependencies: dict[str, set[str]],
                     archive_files: _ArchiveFiles) -> str:
        try:
            spec_output = self.output("spec-export")
        except KeyError:
            return ""
        assert isinstance(spec_output, DirectoryState)
        directory = spec_output.directory
        logging.info("%s: export specification to directory: %s", self.uid,
                     directory)
        member_uids = set(member.uid for member in self.inputs("member"))
        files: list[str] = []
        factory = self.director.factory
        build_order = self.item.view["package-build-order"]
        item_cache = self.director.item_cache
        for uid, item in itertools.chain(item_cache.items(),
                                         item_cache.proxies.items()):
            if item.type.startswith("spec"):
                continue
            data = factory.export_data(item, uid in member_uids, build_order)
            uid_path = f"{uid[1:]}.yml"
            path = os.path.join(directory, uid_path)
            os.makedirs(os.path.dirname(path), exist_ok=True)
            save_data(path, data)
            files.append(uid_path)
        spec_output.set_files(files)
        spec_output.load()
        self._gather_files(spec_output, dependencies, archive_files)
        return ("Export the specification items to directory\n"
                f"{self.description.path(directory)}.")

    def _create_verification_script(self, tar_file: tarfile.TarFile,
                                    strip_prefix: str,
                                    archive_files: _ArchiveFiles) -> str:
        try:
            script_output = self.output("verify-script")
        except KeyError:
            return ""
        assert isinstance(script_output, DirectoryState)
        script_output.set_files((self["verification-script"], ))
        script_path = script_output.file
        logging.info("%s: create the verification script: %s", self.uid,
                     script_path)
        _create_verification_script(script_path, archive_files)
        tar_file.add(script_path, os.path.relpath(script_path, strip_prefix))
        return ("Create the archive verification script\n"
                f"{self.description.path(script_path)}.")

    def run(self) -> None:
        dependencies: dict[str, set[str]] = {}
        archive_files: _ArchiveFiles = {}
        for dir_state in self.inputs("member"):
            assert isinstance(dir_state, DirectoryState)
            self._gather_files(dir_state, dependencies, archive_files)
        export_description = self._export_spec(dependencies, archive_files)

        self.set_files((self["archive-file"], ))
        archive_path = self.file
        logging.info("%s: create archive: %s", self.uid, archive_path)
        os.makedirs(os.path.dirname(archive_path), exist_ok=True)
        with tarfile.open(archive_path, "w:xz") as tar_file:
            strip_prefix = self["archive-strip-prefix"]

            # On some Windows versions, symbolic links have to be created after
            # the link target exists.  Sort the files accordingly.
            dependencies = dict(sorted(dependencies.items()))
            for file_path in graphlib.TopologicalSorter(
                    dependencies).static_order():
                try:
                    archive_files[file_path]
                except KeyError:
                    # This is a symbolic link target outside our control
                    logging.warning(
                        "%s: symbolic link target is "
                        "outside the archive scope: %s", self.uid, file_path)
                    continue
                _check_for_duplicates(self.uid, file_path,
                                      archive_files[file_path])
                tar_file.add(file_path,
                             os.path.relpath(file_path, strip_prefix))

            script_description = self._create_verification_script(
                tar_file, strip_prefix, archive_files)
        logging.info("%s: finished to create archive: %s", self.uid,
                     archive_path)

        self.description.add(f"""Create the archive file
{self.description.path(archive_path)}.
{script_description}
{export_description}""")



# SPDX-License-Identifier: BSD-2-Clause
""" Builds documents. """

# Copyright (C) 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from specitems import Item

from .pkgitems import PackageBuildDirector
from .linkhub import SpecMapper
from .sourcecompare import CompareSourcesProvider
from .speccompare import CompareSpecsProvider
from .sphinxbuilder import SphinxBuilder
from .stdtailoring import StandardTailoringProvider


class DocumentBuilder(SphinxBuilder):
    """ Builds documents. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item,
                         SpecMapper(item["document-key"], self, item))
        self.mapper: SpecMapper
        CompareSourcesProvider(self)
        CompareSpecsProvider(self)
        StandardTailoringProvider(self)



# SPDX-License-Identifier: BSD-2-Clause
""" Aggregates test results. """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import copy
import hashlib
import itertools
import logging
import os
from typing import Any

from specitems import (COL_SPAN, Item, link_is_enabled, make_label,
                       SphinxContent)
from specware import CodeMapper

from .directorystate import DirectoryState
from .pkgitems import BuildItem, PackageBuildDirector
from .rtems import RTEMSItemCache
from .sphinxbuilder import spacify

_Configs = dict[BuildItem, dict[str, list[DirectoryState]]]
_Results = dict[BuildItem, _Configs]
_Data = dict[str, Any]
_SpotToJustification = dict[str, tuple[str, str]]
_FileToSpotJustification = dict[str, _SpotToJustification]


def _add_link(some_data: _Data, report_data: _Data, label: str) -> None:
    some_data["link"] = f"{report_data['report-file']}.html#{label.lower()}"


def _set_limits_and_target_hash(target_data: dict, target: Item) -> None:
    try:
        limits = target.parent("performance-runtime-limits-provider")
    except IndexError as err:
        raise ValueError(
            f"target item {target.uid} has no performance runtime limits"
        ) from err
    assert limits.type == "performance-runtime-limits-provider"
    limits_by_req = {}
    for link in limits.links_to_parents("performance-runtime-limits"):
        limits_by_req[link.item.uid] = link["limits"]
    target_data["limits-by-requirement"] = limits_by_req
    target_data["limits-uid"] = limits.uid
    target_data["target-hash"] = target["target-hash"]


def _test_status(target_data: dict, target: Item, info: dict,
                 verifications: dict[str, str]) -> str:
    link = target_data["link"]
    if info["failed-steps-count"] == 0:
        status = "P"
    elif info.get("uid", None) in verifications:
        status = "X"
    else:
        target.view["no-unexpected-test-failures"] = False
        target.view["validation-status"] = (
            "N/A "
            f"(`at least one unexpected test failure <{link}>`__)")
        return "F"
    target.view.setdefault("no-unexpected-test-failures", True)
    target.view.setdefault("validation-status",
                           f"N/A (`no unexpected test failures <{link}>`__)")
    return status


def get_test_result_status(item: Item,
                           prefix: str = "(",
                           postfix: str = ")") -> str:
    """ Gets the test result status of the item. """
    results: list[str] = []
    for test_results in item.view.get("test-results", {}).values():
        for data in test_results:
            results.append(f"`{data['status']} <{data['link']}>`__")
    if not results:
        return "**no test results**"
    return f"{prefix}{', '.join(results)}{postfix}"


def _update_measurement_status(measurement_data: _Data, env_data: _Data,
                               limits: _Data) -> None:
    if limits["min-lower-bound"] <= env_data[
            "min"] and limits["median-lower-bound"] <= env_data[
                "q2"] <= limits["median-upper-bound"] and env_data[
                    "max"] <= limits["max-upper-bound"]:
        return
    measurement_data["status"] = "F"


def _gather_test_error_verifications(item: Item,
                                     verifications: dict[str, str]) -> None:
    for item_2 in itertools.chain(item.children("verification-member"),
                                  item.parents("verification-provider")):
        _gather_test_error_verifications(item_2, verifications)
    for item_2 in item.parents("verification",
                               is_link_enabled=link_is_enabled):
        assert item_2.uid not in verifications
        verifications[item_2.uid] = item.uid


def _gather_coverage_gap_verifications(
        uid: str, target: Item,
        verifications: dict[str, _FileToSpotJustification]) -> None:
    for child in itertools.chain(target.parents("verification-by"),
                                 target.children("verification")):
        if not child.type.startswith("verification/code-coverage-gap"):
            continue
        for scope_name, scope in child["areas"].items():
            for file_path, file in scope.items():
                logging.info(
                    "%s: for target %s gather code coverage gap "
                    "verifications from %s for file: %s", uid, target.uid,
                    child.uid, file_path)
                gaps: _SpotToJustification = {}
                for line in file["lines"]:
                    line_number = line["line"]
                    logging.info(
                        "%s: add line coverage gap justification at %s", uid,
                        line_number)
                    gaps[f"line/{line_number}"] = child.uid, line["md5"]
                for branches in file["branches"]:
                    line_number = branches["line"]
                    for branch in branches["branches"]:
                        line_branch = f"{line_number}/{branch}"
                        logging.info(
                            "%s: add branch coverage gap justification at %s",
                            uid, line_branch)
                        gaps[f"branch/{line_branch}"] = child.uid, branches[
                            "md5"]
                for function_name in file["functions"]:
                    logging.info(
                        "%s: add function coverage gap justification for %s()",
                        uid, function_name)
                    gaps[f"function/{function_name}"] = child.uid, ""
                verifications.setdefault(scope_name, {})[file_path] = gaps


def _add_coverage_table(content: SphinxContent, none: str, some: str,
                        stats_of_files: list[dict]) -> None:
    if not stats_of_files:
        content.add(none)
        return
    content.add(some)
    rows = [[
        "File", "Functions", "Status", "Lines", "Status", "Branches", "Status"
    ]]
    for stats in stats_of_files:
        row = [stats["file-link"]]
        for kind in _COVERAGE_KINDS:
            row.append(stats[f"{kind}-info"])
            row.append(stats[f"{kind}-status"])
        rows.append(row)
    content.add_grid_table(rows, [34, 13, 7, 13, 7, 13, 7], font_size=-3)


_COVERAGE_KINDS = ("function", "line", "branch")


class _CoverageSummary:
    # pylint: disable=too-many-instance-attributes

    def __init__(self, test_aggregator: "TestAggregator",
                 coverage: dict) -> None:
        self.test_aggregator = test_aggregator
        self.scope = coverage["scope"]
        self.verifications = copy.deepcopy(coverage["verifications"])
        self.html_directory = coverage["html-directory"]
        self.limits_by_area = coverage["limits-by-area"]
        self.good_files: list[dict] = []
        self.justified_files: list[dict] = []
        self.bad_files: list[dict] = []
        self.overall: dict = {}
        self.issues: dict[str, set[str]] = {}
        for kind in _COVERAGE_KINDS:
            self.overall[f"{kind}-covered"] = 0
            self.overall[f"{kind}-justified"] = 0
            self.overall[f"{kind}-total"] = 0
        for file_coverage in coverage["files"]:
            self._add_coverage_of_file(file_coverage)
        limits = self.limits_by_area["overall"]
        for kind in _COVERAGE_KINDS:
            self.overall[f"{kind}-justified-covered"] = self.overall[
                f"{kind}-covered"]
            self._add_coverage_status(self.overall,
                                      limits[f"{kind}-min-percent"], True,
                                      kind)
        if self.verifications:
            unused: list[str] = sorted(
                set(justification[0]
                    for spot_to_justification in self.verifications.values()
                    for justification in spot_to_justification.values()))
            logging.warning(
                "%s: for %s there are unused code coverage justifications: %s",
                self.test_aggregator.uid, coverage["target-uid"], unused)
            self.issues.setdefault("Unused code coverage justifications",
                                   set()).update(f"spec:{spacify(uid)}"
                                                 for uid in unused)

    def get_issues(self, issues: dict[str, set[str]]) -> None:
        """ Get the detected coverage issues. """
        for key, items in self.issues.items():
            issues.setdefault(key, set()).update(items)

    def add_coverage_section(self, content: SphinxContent) -> None:
        """ Add a coverage section to the content. """
        with content.section(f"Scope - {self.scope}"):
            if any((self.good_files, self.justified_files, self.bad_files)):
                _add_coverage_table(
                    content,
                    "There are no files with unjustified coverage issues.",
                    "The following table lists files "
                    "with unjustified coverage issues.", self.bad_files)
                _add_coverage_table(
                    content,
                    "There are no files with justified coverage issues.",
                    "The following table lists files "
                    "with justified coverage issues.", self.justified_files)
                _add_coverage_table(
                    content, "There are no files without coverage issues.",
                    "The following table lists files "
                    "having the expected coverage.", self.good_files)
            else:
                content.add("There is no coverage information available.")

    def _add_coverage_status(self, stats: dict, limit: float, overall: bool,
                             kind: str) -> None:
        total = stats[f"{kind}-total"]
        if not total:
            stats[f"{kind}-info"] = "N/A"
            if overall:
                stats[f"{kind}-status"] = "**NOK**"
                stats[f"{kind}-error"] = True
                stats["error"] = True
                self.issues.setdefault(
                    f"No {kind} information in coverage data",
                    set()).add(f"Scope - {self.scope}")
            else:
                stats[f"{kind}-status"] = "OK"
            return

        covered = stats[f"{kind}-covered"]
        justified = stats[f"{kind}-justified"]
        if justified != 0:
            info = f"{covered}+{justified}/{total}"
        else:
            info = f"{covered}/{total}"
        both = covered + justified
        if both == total:
            stats[f"{kind}-info"] = f"{info} (100%)"
            stats[f"{kind}-status"] = "OK"
            return

        # Make sure we never display 100.0% if not everything is covered
        percent = float((1000 * both) // total) / 10.0

        if percent >= limit:
            status = "OK"
        else:
            status = "**NOK**"
            stats[f"{kind}-error"] = True
            stats["error"] = True
            if overall:
                self.overall[f"{kind}-error"] = True
                self.issues.setdefault(f"Insufficient overall {kind} coverage",
                                       set()).add(f"Scope - {self.scope}")
            else:
                self.issues.setdefault(
                    f"Insufficient file-specific {kind} coverage",
                    set()).add(stats["file-link"])
        stats[f"{kind}-info"] = f"{info} ({percent:.1f}%)"
        stats[f"{kind}-status"] = status

    def _add_file_stats(self, stats: dict[str, int | str], file_path: str,
                        spot_to_justification: _SpotToJustification) -> None:
        limits = self.limits_by_area.get(f"file-{file_path}",
                                         self.limits_by_area["per-file"])
        for kind in _COVERAGE_KINDS:
            for what in ("covered", "justified", "total"):
                key = f"{kind}-{what}"
                self.overall[key] += stats[key]
            self._add_coverage_status(stats, limits[f"{kind}-min-percent"],
                                      False, kind)
        if spot_to_justification:
            unrelated: list[str] = sorted(
                set(justification[0]
                    for justification in spot_to_justification.values()))
            logging.warning(
                "%s: for file %s there are unrelated "
                "code coverage justifications: %s", self.test_aggregator.uid,
                file_path, unrelated)
            self.issues.setdefault("Unrelated code coverage justifications",
                                   set()).update(f"spec:{spacify(uid)}"
                                                 for uid in unrelated)
        if "error" in stats:
            self.bad_files.append(stats)
        elif "justified" in stats:
            self.justified_files.append(stats)
        else:
            self.good_files.append(stats)

    def _add_line_stats(self, stats: dict, line: dict,
                        spot_to_justification: _SpotToJustification) -> None:
        if line["count"] > 0:
            stats["line-covered"] += 1
        else:
            file_path = stats["file-path"]
            line_no = line["line_number"]
            justification = spot_to_justification.pop(f"line/{line_no}", None)
            if justification is None:
                logging.info(
                    "%s: no line coverage gap justification for %s:%s",
                    self.test_aggregator.uid, file_path, line_no)
            else:
                uid = justification[0]
                logging.info(
                    "%s: use line coverage gap "
                    "justification %s for %s:%s", self.test_aggregator.uid,
                    uid, file_path, line_no)
                if line["gcovr/md5"] == justification[1]:
                    stats["justified"] = True
                    stats["line-justified"] += 1
                else:
                    logging.warning(
                        "%s: out of date line coverage gap "
                        "justification %s for %s:%s", self.test_aggregator.uid,
                        uid, file_path, line_no)
                    self.issues.setdefault(
                        "Out of date line coverage gap justifications",
                        set()).add(f"spec:{spacify(uid)}")

    def _add_branch_stats(self, stats: dict, line: dict, branch: dict,
                          spot_to_justification: _SpotToJustification) -> None:
        stats["branch-total"] += 1
        if branch["count"] > 0:
            stats["branch-covered"] += 1
        else:
            file_path = stats["file-path"]

            # Since GCC 14, the JSON gcov output is used by gcovr.  In this
            # case, branches are identified by a "source_block_id" and
            # "destination_block_id" pair.  The "source_block_id" is the same
            # for all branches of a line.  In GCC 13 or earlier, a "branchno"
            # is used.  This means that the branch identification depends on
            # the target GCC version.
            branchno = branch.get("destination_block_id",
                                  branch.get("branchno", -1))

            line_branch = f"{line['line_number']}/{branchno}"
            justification = spot_to_justification.pop(f"branch/{line_branch}",
                                                      None)
            if justification is None:
                logging.info(
                    "%s: no branch coverage gap justification for %s:%s",
                    self.test_aggregator.uid, file_path, line_branch)
            else:
                uid = justification[0]
                logging.info(
                    "%s: use branch coverage gap "
                    "justification %s for %s:%s", self.test_aggregator.uid,
                    uid, file_path, line_branch)
                if line["gcovr/md5"] == justification[1]:
                    stats["justified"] = True
                    stats["branch-justified"] += 1
                else:
                    logging.warning(
                        "%s: out of date branch coverage gap "
                        "justification %s for %s:%s", self.test_aggregator.uid,
                        uid, file_path, line_branch)
                    self.issues.setdefault(
                        "Out of date branch coverage gap justifications",
                        set()).add(f"spec:{spacify(uid)}")

    def _add_function_stats(
            self, stats: dict, function: dict,
            spot_to_justification: _SpotToJustification) -> None:
        execution_count = function.get("execution_count", None)
        if execution_count is None:
            # Sometimes gcov is unable to find an associated name for some
            # function code blocks.  This results in function data with a name
            # of "<unknown function N>" and no execution count.  The code
            # blocks still have line and branch coverage data.
            return
        stats["function-total"] += 1
        if execution_count > 0:
            stats["function-covered"] += 1
        else:
            file_path = stats["file-path"]
            function_name = function["name"]
            justification = spot_to_justification.pop(
                f"function/{function_name}", None)
            if justification is None:
                logging.info(
                    "%s: no function coverage gap "
                    "justification for %s() in %s", self.test_aggregator.uid,
                    function_name, file_path)
            else:
                uid = justification[0]
                logging.info(
                    "%s: use function coverage gap "
                    "justification %s for %s() in %s",
                    self.test_aggregator.uid, uid, function_name, file_path)
                stats["justified"] = True
                stats["function-justified"] += 1

    def _add_coverage_of_file(self, file_coverage: dict) -> None:
        file_path = file_coverage["file"]
        digest = hashlib.md5(file_path.encode("utf-8"),
                             usedforsecurity=False).hexdigest()
        link = f"{os.path.basename(file_path)}.{digest}"
        link = f"{self.html_directory}/index.{link}.html"
        file_link = f"`{spacify(file_path)} <{link}>`__"
        spot_to_justification = self.verifications.pop(file_path, {})
        stats = {
            "file-path": file_path,
            "file-link": file_link,
            "branch-covered": 0,
            "branch-justified": 0,
            "branch-total": 0,
            "function-covered": 0,
            "function-justified": 0,
            "function-total": 0,
            "line-covered": 0,
            "line-justified": 0,
            "line-total": 0
        }
        for line in file_coverage["lines"]:
            stats["line-total"] += 1
            self._add_line_stats(stats, line, spot_to_justification)
            for branch in line["branches"]:
                self._add_branch_stats(stats, line, branch,
                                       spot_to_justification)
        for function in file_coverage["functions"]:
            self._add_function_stats(stats, function, spot_to_justification)
        self._add_file_stats(stats, file_path, spot_to_justification)


class TestAggregator(BuildItem):
    """ Aggregates test results. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.report_file: str = self["report-file"]
        self.report_directory = os.path.dirname(self.report_file)
        self.targets: dict[str, dict] = {}
        self.runtime_measurements: list[_Data] = []
        spec = self.input("spec")
        assert isinstance(spec, RTEMSItemCache)
        self.spec = spec
        for target, configs in sorted(self._gather_results().items()):
            target_key = target["key"]
            label = self._make_label({}, target_key)
            test_error_verifications: dict[str, str] = {}
            _gather_test_error_verifications(target.item,
                                             test_error_verifications)
            coverage_gap_verifications: dict[str,
                                             _FileToSpotJustification] = {}
            _gather_coverage_gap_verifications(self.uid, target.item,
                                               coverage_gap_verifications)
            target_data = {
                "configs": [],
                "key": target_key,
                "label": label,
                "link": self._make_report_link(label),
                "name": CodeMapper(target.item).substitute(target["name"]),
                "uid": target.uid,
                "test-error-verifications": test_error_verifications,
                "coverage-gap-verifications": coverage_gap_verifications
            }
            _set_limits_and_target_hash(target_data, target.item)
            logging.debug("%s: add target %s with hash %s", self.uid,
                          target_data["uid"], target_data["target-hash"])
            self._add_configs_to_target(spec, configs, target_data, target_key)
            self.targets[target.uid] = target_data
        spec.validate_using_test_results()

    def run(self) -> None:
        self.description.add("Aggregate test results.")

    def _add_configs_to_target(self, spec: RTEMSItemCache, configs: _Configs,
                               target_data: dict, target_key: str) -> None:
        for config, results in sorted(configs.items()):
            config_key = config["config-key"]
            label = self._make_label(target_data, config_key)
            config_data = {
                "build-label": config["build-label"],
                "config-key": config_key,
                "description": config["description"],
                "label": label,
                "link": self._make_report_link(label),
                "name": config["name"],
                "other-programs": {},
                "report-file-base": f"{target_key}-{config_key}",
                "target": target_data,
                "test-programs": {},
                "test-suites": {},
            }
            logging.debug("%s: add configuration %s to target %s", self.uid,
                          config_key, target_data["uid"])
            for test_log in results.get("test-log", []):
                self._add_test_reports(spec, test_log, config_data)
            for coverage in results.get("test-coverage", []):
                coverage_data = coverage.json_load()
                scope = coverage["scope"]
                coverage_data["scope"] = scope
                coverage_data["target-uid"] = target_data["uid"]
                coverage_data["limits-by-area"] = coverage["limits"]
                coverage_data["html-directory"] = coverage.output(
                    "html")["directory"]
                coverage_data["verifications"] = target_data[
                    "coverage-gap-verifications"].get(scope, {})
                config_data.setdefault("coverage", []).append(coverage_data)
            target_data["configs"].append(config_data)

    def _make_label(self, data: dict, name: str) -> str:
        return f"{data.get('label', '')}{make_label(name)}"

    def _make_report_link(self, label: str) -> str:
        return f"{self.report_file}#{label.lower()}"

    def _make_report_file(self, config_data: _Data, uid: str) -> str:
        base = config_data["report-file-base"]
        name = uid.replace("/", "-")
        return f"{self.report_directory}/{base}{name}"

    def _gather_results(self) -> _Results:
        results: _Results = {}
        for link, file_state in itertools.chain(
                self.input_links("test-log"),
                self.input_links("test-coverage")):
            assert isinstance(file_state, DirectoryState)
            target = file_state.input("target")
            build_config = file_state.input("build-configuration")
            results.setdefault(target,
                               {}).setdefault(build_config, {}).setdefault(
                                   link["name"], []).append(file_state)
        return results

    def _add_test_reports(self, spec: RTEMSItemCache, test_log: DirectoryState,
                          config_data: _Data) -> None:
        logging.debug("%s: process %s", self.uid, test_log.file)
        test_log_data = test_log.json_load()
        description = test_log_data.get("test-runner-description", "")
        target_data = config_data["target"]
        if "test-runner-description" in target_data:
            assert not description or target_data[
                "test-runner-description"] == description
        else:
            target_data["test-runner-description"] = description
        for report in test_log_data["reports"]:
            try:
                test_suite = report["test-suite"]
                test_suite_item = spec.name_to_item[test_suite["name"]]
            except KeyError:
                executable = os.path.basename(report["executable"])
                name = f"file/{executable}"
                try:
                    uid = spec.name_to_item[name].uid
                except KeyError:
                    logging.debug("%s: add other program %s", self.uid,
                                  executable)
                    assert executable not in config_data["other-programs"]
                    report["report-file"] = self._make_report_file(
                        config_data, f"-{executable}")
                    config_data["other-programs"][executable] = report
                else:
                    logging.debug("%s: add test program %s", self.uid, uid)
                    assert uid not in config_data["test-programs"]
                    report["report-file"] = self._make_report_file(
                        config_data, uid)
                    config_data["test-programs"][uid] = report
            else:
                self._process_test_suite(spec, report, config_data, test_suite,
                                         test_suite_item)

    def _process_runtime_measurements(self, spec: RTEMSItemCache,
                                      runtime_measurements: _Data,
                                      config_data: _Data, test_suite: _Data,
                                      test_case: _Data) -> None:
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments
        # pylint: disable=too-many-locals
        target_data = config_data["target"]
        target_uid = target_data["uid"]
        limits_by_req = target_data["limits-by-requirement"]
        for env_data in test_case["runtime-measurements"]:
            req = spec.name_to_item[env_data["name"]]
            env_data["requirement-uid"] = req.uid
            env_data["test-case"] = test_case
            measurement_data = runtime_measurements.get(req.uid, None)
            if measurement_data is None:
                label = self._make_label(test_suite, req.uid)
                measurement_data = {
                    "config": config_data,
                    "label": label,
                    "requirement-uid": req.uid,
                    "status": "P",
                    "test-case": test_case,
                    "test-suite": test_suite,
                    "variants": {}
                }
                _add_link(measurement_data, test_suite, label)
                runtime_measurements[req.uid] = measurement_data
                req.view.setdefault("test-results",
                                    {}).setdefault(target_uid,
                                                   []).append(measurement_data)
            else:
                test_case_2 = measurement_data["test-case"]
                assert test_case_2["name"] == test_case["name"]
                assert test_case_2["test-suite"]["name"] == test_suite["name"]
            env_name = env_data["variant"]
            measurement_data["variants"][env_name] = env_data
            label = self._make_label(measurement_data, env_name)
            env_data["label"] = label
            _add_link(env_data, test_suite, label)
            _update_measurement_status(measurement_data, env_data,
                                       limits_by_req[req.uid][env_name])
            self.runtime_measurements.append(measurement_data)

    def _process_test_suite(self, spec: RTEMSItemCache, report: Any,
                            config_data: _Data, test_suite: dict[str, Any],
                            test_suite_item: Item) -> None:
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments
        # pylint: disable=too-many-locals
        logging.debug("%s: process test suite %s", self.uid,
                      test_suite_item.uid)
        target_data = config_data["target"]
        target_uid = target_data["uid"]
        target_item = self.item.cache[target_uid]
        verifications = target_data["test-error-verifications"]
        test_suite["uid"] = test_suite_item.uid
        runtime_measurements: _Data = {}
        test_cases: dict = {}
        unspec_test_cases: list = []
        assert test_suite_item.uid not in config_data["test-suites"]
        label = self._make_label(config_data, test_suite["name"])
        report_file = self._make_report_file(config_data, test_suite_item.uid)
        test_suite_data = {
            "config":
            config_data,
            "label":
            label,
            "report-file":
            report_file,
            "report":
            report,
            "status":
            _test_status(target_data, target_item, test_suite, verifications),
            "runtime-measurements":
            runtime_measurements,
            "test-cases":
            test_cases,
            "unspecified-test-cases":
            unspec_test_cases
        }
        _add_link(test_suite_data, test_suite_data, label)
        config_data["test-suites"][test_suite_item.uid] = test_suite_data
        test_suite_item.view.setdefault("test-results", {}).setdefault(
            target_uid, []).append(test_suite_data)
        test_suite["config"] = config_data
        test_suite["label"] = label
        test_suite["report-file"] = report_file
        _add_link(test_suite, test_suite_data, label)
        for test_case in test_suite["test-cases"]:
            test_case_name = test_case["name"]
            label = self._make_label(test_suite, test_case_name)
            test_case["config"] = config_data
            test_case["label"] = label
            test_case["test-suite"] = test_suite
            _add_link(test_case, test_suite, label)
            no_spec_for_test_case = False
            try:
                test_case_item = spec.name_to_item[test_case_name]
            except KeyError:
                logging.debug("%s: test case without specification %s",
                              self.uid, test_case_name)
                no_spec_for_test_case = True
            else:
                test_case["uid"] = test_case_item.uid
                test_cases[test_case_item.uid] = test_case
                test_case_item.view.setdefault("test-results", {}).setdefault(
                    target_uid, []).append(test_case)
                self._process_runtime_measurements(spec, runtime_measurements,
                                                   config_data, test_suite,
                                                   test_case)
            test_case["status"] = _test_status(target_data, target_item,
                                               test_case, verifications)
            for remark in test_case["remarks"]:
                try:
                    remark_item = spec.name_to_item[remark["remark"]]
                except KeyError:
                    no_spec_for_test_case = True
                else:
                    remark["uid"] = remark_item.uid
                    remark_item.view.setdefault("test-results", {}).setdefault(
                        target_uid, []).append(test_case)
            if no_spec_for_test_case:
                unspec_test_cases.append(test_case)

    def add_coverage_achievement(self, content: SphinxContent) -> None:
        """ Add the code/branch coverage achievement to the content. """
        rows: list[list[str | int]] = [[
            "Target", "Configuration", "Scope", "Functions", "Status", "Lines",
            "Status", "Branches", "Status"
        ]]
        for target_data in self.targets.values():
            target: str | int = (f"`{target_data['name']} "
                                 f"<{target_data['link']}>`__")
            for config_data in target_data["configs"]:
                key = config_data["config-key"]
                config: str | int = f"`{key} <{config_data['link']}>`__"
                for coverage in config_data.get("coverage", []):
                    summary = _CoverageSummary(self, coverage)
                    row = [target, config, coverage["scope"]]
                    for kind in _COVERAGE_KINDS:
                        row.append(summary.overall[f"{kind}-info"])
                        row.append(summary.overall[f"{kind}-status"])
                    rows.append(row)
                    config = COL_SPAN
                    target = COL_SPAN
        content.add_grid_table(rows, [18, 8, 8, 13, 7, 13, 7, 13, 7],
                               font_size=-3)

    def add_coverage_of_config(self, content: SphinxContent,
                               config_data: _Data,
                               issues: dict[str, set[str]]) -> None:
        """
        Add the code/branch coverage data associated with the configuration
        data to the content.

        Get the detected code coverage issues.
        """
        summaries: list[_CoverageSummary] = []
        with content.section("Overview"):
            rows = [[
                "Scope", "Functions", "Status", "Lines", "Status", "Branches",
                "Status"
            ]]
            for coverage in config_data.get("coverage", []):
                summary = _CoverageSummary(self, coverage)
                summaries.append(summary)
                summary.get_issues(issues)
                row = [summary.scope]
                for kind in _COVERAGE_KINDS:
                    row.append(summary.overall[f"{kind}-info"])
                    row.append(summary.overall[f"{kind}-status"])
                rows.append(row)
            content.add_grid_table(rows, [34, 13, 7, 13, 7, 13, 7],
                                   font_size=-3)
        for summary in summaries:
            summary.add_coverage_section(content)
